# pylint: disable=global-variable-not-assigned
# pylint: disable=missing-function-docstring
"""
Environment variable and settings management. See the
[README](https://cedric.hothanh.fr/turbo-broccoli/turbo_broccoli.html#environment-variables)
for information about the supported environment variables.
"""
__docformat__ = "google"

try:
    from loguru import logger as logging
except ModuleNotFoundError:
    import logging  # type: ignore

import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

_DATACLASSES_TYPES: Dict[str, type] = {}
_PYTORCH_MODULE_TYPES: Dict[str, type] = {}

# The initial values are the defaults
_ENVIRONMENT: Dict[str, Any] = {
    "TB_ARTIFACT_PATH": Path("./"),
    "TB_KERAS_FORMAT": "tf",
    "TB_MAX_NBYTES": 8_000,
    "TB_NODECODE": [],
    "TB_PANDAS_FORMAT": "h5",
    "TB_SHARED_KEY": None,
}


def _init():
    """
    Reads the environment and sets the
    `turbo_broccoli.environment._ENVIRONMENT` accordingly.
    """
    if "TB_NUMPY_PATH" in os.environ:
        logging.warning(
            "The use of the 'TB_NUMPY_PATH' environment variable is "
            "deprecated. Consider using 'TB_ARTIFACT_PATH' instead."
        )
        set_artifact_path(Path(os.environ["TB_NUMPY_PATH"]))
    else:
        set_artifact_path(
            os.environ.get(
                "TB_ARTIFACT_PATH",
                _ENVIRONMENT["TB_ARTIFACT_PATH"],
            )
        )

    try:
        set_keras_format(
            os.environ.get(
                "TB_KERAS_FORMAT",
                _ENVIRONMENT["TB_KERAS_FORMAT"],
            )
        )
    except ValueError:
        logging.warning(
            "Invalid value for environment variable TB_KERAS_FORMAT: "
            f"'{os.environ['TB_KERAS_FORMAT']}'. Expected 'h5', 'json', or "
            f"'tf'. Defaulting to '{_ENVIRONMENT['TB_KERAS_FORMAT']}'."
        )

    try:
        set_pandas_format(
            os.environ.get(
                "TB_PANDAS_FORMAT",
                _ENVIRONMENT["TB_PANDAS_FORMAT"],
            )
        )
    except ValueError:
        logging.warning(
            "Invalid value for environment variable TB_PANDAS_FORMAT: "
            f"'{os.environ['TB_PANDAS_FORMAT']}'. Expected 'csv', 'excel', "
            "'feather', 'h5', 'hdf', 'pickle', 'stata', or 'xml'. Defaulting "
            f"to '{_ENVIRONMENT['TB_PANDAS_FORMAT']}'."
        )

    if "TB_NODECODE" in os.environ:
        set_nodecode(os.environ["TB_NODECODE"])

    if "TB_NUMPY_MAX_NBYTES" in os.environ:
        logging.warning(
            "The use of the 'TB_NUMPY_MAX_NBYTES' environment variable is "
            "deprecated. Consider using 'TB_MAX_NBYTES' instead."
        )
        set_max_nbytes(int(os.environ["TB_NUMPY_MAX_NBYTES"]))
    else:
        set_max_nbytes(
            int(
                os.environ.get(
                    "TB_MAX_NBYTES",
                    _ENVIRONMENT["TB_MAX_NBYTES"],
                )
            )
        )

    if "TB_SHARED_KEY" in os.environ:
        set_shared_key(os.environ["TB_SHARED_KEY"])


def get_artifact_path() -> Path:
    return _ENVIRONMENT["TB_ARTIFACT_PATH"]


def get_keras_format() -> str:
    return _ENVIRONMENT["TB_KERAS_FORMAT"]


def get_pandas_format() -> str:
    return _ENVIRONMENT["TB_PANDAS_FORMAT"]


def get_max_nbytes() -> int:
    return _ENVIRONMENT["TB_MAX_NBYTES"]


def get_registered_dataclass_type(name: str) -> type:
    return _DATACLASSES_TYPES[name]


def get_registered_pytorch_module_type(name: str) -> type:
    return _PYTORCH_MODULE_TYPES[name]


def get_shared_key() -> Optional[bytes]:
    return _ENVIRONMENT["TB_SHARED_KEY"]


def is_nodecode(type_name: str) -> bool:
    return type_name in _ENVIRONMENT["TB_NODECODE"]


def register_dataclass_type(cls: type):
    """
    Registers a dataclass for dataclass deserialization. Registered types may
    be overwritten.
    """
    _DATACLASSES_TYPES[cls.__name__] = cls


def register_pytorch_module_type(cls: type):
    """
    Registers a `torch.nn.Module` type for module deserialization. Registered
    types may be overwritten.
    """
    _PYTORCH_MODULE_TYPES[cls.__name__] = cls


def set_artifact_path(path: Union[str, Path]):
    if isinstance(path, str):
        path = Path(path)
    if not (path.exists() and path.is_dir()):
        raise RuntimeError(
            f"Path {str(path)} does not point to an existing directory"
        )
    _ENVIRONMENT["TB_ARTIFACT_PATH"] = path


def set_keras_format(fmt: str):
    """Valid format are `h5`, `json`, and `tf`."""
    fmt = fmt.lower()
    KERAS_FORMATS = ["h5", "json", "tf"]
    if fmt not in KERAS_FORMATS:
        raise ValueError(
            f"Invalid value for environment variable TB_KERAS_FORMAT: {fmt}. "
            "Valid formats are 'h5', 'json', and 'tf' (the default)."
        )
    _ENVIRONMENT["TB_KERAS_FORMAT"] = fmt


def set_max_nbytes(nbytes: int):
    if nbytes <= 0:
        raise ValueError("numpy's max nbytes must be > 0")
    _ENVIRONMENT["TB_MAX_NBYTES"] = nbytes


def set_nodecode(types: Union[str, List[str]]):
    _ENVIRONMENT["TB_NODECODE"] = (
        types.split(",") if isinstance(types, str) else types
    )


def set_pandas_format(fmt: str):
    """
    Valid formats are `csv`, `excel`, `feather`, `h5`, `hdf`, `html`, `pickle`,
    `sql`, `stata`, and `xml`.

    TODO: Write all the unit tests.
    """
    fmt = fmt.lower()
    PANDAS_FORMATS = [
        "csv",
        "excel",
        "feather",
        "h5",
        "hdf",
        "html",
        "pickle",
        "sql",
        "stata",
        "xml",
    ]
    if fmt not in PANDAS_FORMATS:
        raise ValueError(
            f"Invalid value for environment variable TB_PANDAS_FORMAT: {fmt}. "
            "Valid formats are `csv`, `excel`, `feather`, `h5` (the default), "
            "`hdf`, `html`, `pickle`, `sql`, `stata`, and `xml`."
        )
    _ENVIRONMENT["TB_PANDAS_FORMAT"] = fmt


def set_shared_key(key: Optional[Union[str, bytes]]):
    """If the provided key is a string, it will be encoded in `utf-8`."""
    if isinstance(key, str):
        key = key.encode("utf-8")
    _ENVIRONMENT["TB_SHARED_KEY"] = key


_init()



"""
Serialization of so-called generic object. See
`turbo_broccoli.generic.to_json`.
"""
__docformat__ = "google"


from typing import Any, Iterable


def to_json(obj: Any) -> dict:
    """Serializes a generic object into JSON by cases."""
    if not (
        hasattr(obj, "__turbo_broccoli__")
        and isinstance(obj.__turbo_broccoli__, Iterable)
    ):
        raise TypeError("Not a generic object")
    return {
        "__generic__": {
            "__version__": 1,
            "data": {k: getattr(obj, k) for k in obj.__turbo_broccoli__},
        },
    }



"""Guarded call"""

from pathlib import Path
import hashlib
from functools import partial

try:
    from loguru import logger as logging
except ModuleNotFoundError:
    import logging  # type: ignore

import json
from typing import Any, Callable, Dict, Union

from .turbo_broccoli import TurboBroccoliDecoder, TurboBroccoliEncoder, to_json


def guarded_call(
    function: Callable[..., Dict[str, Any]],
    path: Union[str, Path],
    *args,
    **kwargs,
) -> Dict[str, Any]:
    """
    Convenience function:

    ```py
    guarded_call(f, "out/result.json", *args, **kwargs)
    ```
    is equivalent to

    ```py
    _f = produces_document(f, "out/result.json")
    _f(*args, **kwargs)
    ```
    """
    _f = produces_document(function, path)
    return _f(*args, **kwargs)


def produces_document(
    function: Callable[..., Dict[str, Any]],
    path: Union[str, Path],
    check_args: bool = False,
) -> Callable[..., Dict[str, Any]]:
    """
    Consider an expensive function `f` that returns a TurboBroccoli/JSON-izable
    `dict`. Wrapping/decorating `f` using `produces_document` essentially saves
    the result at a specified path and when possible, loads it instead of
    calling `f`. For example:

    ```py
    _f = produces_document(f, "out/result.json")
    _f(*args, **kwargs)
    ```

    will only call `f` if the `out/result.json` does not exist, and otherwise,
    loads and returns `out/result.json`. However, if `out/result.json` exists
    and was produced by calling `_f(*args, **kwargs)`, then

    ```py
    _f(*args2, **kwargs2)
    ```

    will still return the same result. If you want to keep a different file for
    each `args`/`kwargs`, set `check_args` to `True` as in

    ```py
    _f = produces_document(f, "out/result.json")
    _f(*args, **kwargs)
    ```

    In this case, the arguments must be TurboBroccoli/JSON-izable, i.e. the
    document

    ```
    {
        "args": args,
        "kwargs": kwargs,
    }
    ```

    must be TurboBroccoli/JSON-izable. The resulting file is no longer
    `out/result.json` but rather `out/result.json/<hash>` where `hash` is the
    MD5 hash of the serialization of the `args`/`kwargs` document above.
    """

    def _wrapped(path: Path, *args, **kwargs) -> Dict[str, Any]:
        if check_args:
            s = to_json({"args": args, "kwargs": kwargs})
            h = hashlib.md5(s.encode("utf-8")).hexdigest()
            path.mkdir(parents=True, exist_ok=True)
            path = path / f"{h}.json"
        try:
            with open(path, "r", encoding="utf-8") as fp:
                obj = json.load(fp, cls=TurboBroccoliDecoder)
            logging.debug(
                f"Skipped call to guarded method '{function.__name__}'"
            )
            return obj
        except:  # pylint: disable=bare-except
            obj = function(*args, **kwargs)
            with open(path, "w", encoding="utf-8") as fp:
                json.dump(obj, fp, cls=TurboBroccoliEncoder)
            return obj

    path = path if isinstance(path, Path) else Path(path)
    return partial(_wrapped, path)



"""Dataclass serialization"""
__docformat__ = "google"

from typing import Any

from turbo_broccoli.environment import (
    get_registered_dataclass_type,
    is_nodecode,
)


def _json_to_dataclass_v2(dct: dict) -> Any:
    """
    Converts a JSON document following the v2 specification to a dataclass
    object.
    """
    if is_nodecode("dataclass." + dct["class"]):
        return None
    return get_registered_dataclass_type(dct["class"])(**dct["data"])


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a dataclass object. See `to_json` for the
    specification `dct` is expected to follow, and
    `turbo_broccoli.environment.register_dataclass_type`.
    """
    DECODERS = {
        2: _json_to_dataclass_v2,
    }
    try:
        return DECODERS[dct["__dataclass__"]["__version__"]](
            dct["__dataclass__"]
        )
    except KeyError as exc:
        raise TypeError("Not a valid dataclass document") from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a dataclass into JSON by cases. The return dict has the
    following structure

        {
            "__dataclass__": {
                "__version__": 2,
                "class": <str>,
                "data": {...},
            },
        }

    where the `{...}` is `obj.__dict__`.
    """
    if hasattr(obj, "__dataclass_fields__"):
        return {
            "__dataclass__": {
                "__version__": 2,
                "class": obj.__class__.__name__,
                "data": obj.__dict__,
            },
        }
    raise TypeError("Not a supported type")



"""Various utilities and internal methods"""

try:
    from loguru import logger as logging
except ModuleNotFoundError:
    import logging  # type: ignore

_WARNED_ABOUT_SAFETENSORS = False


def warn_about_safetensors():
    """
    If safetensors is not installed, logs a warning message. This method may be
    called multiple times, but the message will only be logged once.
    """
    global _WARNED_ABOUT_SAFETENSORS  # pylint: disable=global-statement
    if not _WARNED_ABOUT_SAFETENSORS:
        logging.warning(
            "Serialization of numpy arrays and Tensorflow tensors without "
            "safetensors is deprecated. Consider installing safetensors using "
            "'pip install safetensors'."
        )
        _WARNED_ABOUT_SAFETENSORS = True



# pylint: disable=missing-function-docstring
"""Test suite for `tdt.produces_document`"""

import json
from turbo_broccoli import guarded_call, produces_document


def test_guarded_call():
    def f(a: int):
        return {"a": a}

    path = "out/test_guarded_call.json"
    x = guarded_call(f, path, 1)
    with open(path, "r", encoding="utf-8") as fp:
        y = json.load(fp)
    assert isinstance(x, dict)
    assert x == y
    assert x != f(2)
    assert x == guarded_call(f, path, 2)  # Intended behavior


def test_produces_document():
    def f(a: int):
        return {"a": a}

    path = "out/test_produces_document.json"
    _f = produces_document(f, path, check_args=False)
    x = _f(1)
    with open(path, "r", encoding="utf-8") as fp:
        y = json.load(fp)
    assert isinstance(x, dict)
    assert x == y
    assert x != f(2)
    assert x == _f(2)  # Intended behavior


def test_produces_document_check_args():
    def f(a: int):
        return {"a": a}

    path = "out/test_produces_document_check_args.json"
    _f = produces_document(f, path, check_args=True)
    assert _f(1) == {"a": 1}
    assert _f(2) == {"a": 2}
    assert _f(1) == {"a": 1}  # Repetition intended
    assert _f(2) == {"a": 2}  # Repetition intended
    assert _f(1) != _f(2)



"""
.. include:: ../README.md
.. include:: ../CHANGELOG.md
"""
__docformat__ = "google"

from pkg_resources import DistributionNotFound, get_distribution

from turbo_broccoli.environment import (
    register_dataclass_type,
    register_pytorch_module_type,
    set_artifact_path,
    set_keras_format,
    set_max_nbytes,
    set_pandas_format,
)
from turbo_broccoli.guard import guarded_call, produces_document
from turbo_broccoli.turbo_broccoli import (
    TurboBroccoliDecoder,
    TurboBroccoliEncoder,
    from_json,
    to_json,
)

try:
    __version__ = get_distribution("turbo-broccoli").version
except DistributionNotFound:
    __version__ = "local"



# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=unused-wildcard-import
# pylint: disable=wildcard-import
"""sklearn estimators (de)serialization test suite"""

from typing import Optional, Tuple

import numpy as np
from common import from_json, to_json
from numpy.testing import assert_array_equal
from scipy import sparse
from sklearn.calibration import *
from sklearn.cluster import *
from sklearn.compose import *
from sklearn.covariance import *
from sklearn.cross_decomposition import *
from sklearn.datasets import *
from sklearn.datasets import load_digits, make_sparse_coded_signal
from sklearn.decomposition import *
from sklearn.discriminant_analysis import *
from sklearn.ensemble import *
from sklearn.exceptions import *
from sklearn.feature_extraction import *
from sklearn.feature_selection import *
from sklearn.gaussian_process import *
from sklearn.gaussian_process.kernels import RBF, DotProduct, WhiteKernel
from sklearn.impute import *
from sklearn.inspection import *
from sklearn.isotonic import *
from sklearn.kernel_approximation import *
from sklearn.kernel_ridge import *
from sklearn.linear_model import *
from sklearn.manifold import *
from sklearn.metrics import *
from sklearn.mixture import *
from sklearn.model_selection import *
from sklearn.multiclass import *
from sklearn.multioutput import *
from sklearn.naive_bayes import *
from sklearn.neighbors import *
from sklearn.neural_network import *
from sklearn.pipeline import *
from sklearn.preprocessing import *
from sklearn.random_projection import *
from sklearn.semi_supervised import *
from sklearn.svm import *
from sklearn.tree import *
from sklearn.metrics.pairwise import pairwise_kernels


def _to_json_and_back(obj: BaseEstimator) -> BaseEstimator:
    return from_json(to_json({"e": obj}))["e"]


def _fit_labels_test(
    obj: BaseEstimator, x: np.ndarray
) -> Tuple[BaseEstimator, BaseEstimator]:
    obj = obj.fit(x)
    obj2 = _to_json_and_back(obj)
    assert_array_equal(obj.labels_, obj.labels_)
    return obj, obj2


def _fit_predict_x_z_test(
    obj: BaseEstimator, x: np.ndarray, z: np.ndarray
) -> Tuple[BaseEstimator, BaseEstimator]:
    obj = obj.fit(x)
    obj2 = _to_json_and_back(obj)
    assert_array_equal(obj.predict(z), obj2.predict(z))
    return obj, obj2


def _fit_predict_x_y_z_test(
    obj: BaseEstimator, x: np.ndarray, y: np.ndarray, z: np.ndarray
) -> Tuple[BaseEstimator, BaseEstimator]:
    obj = obj.fit(x, y)
    obj2 = _to_json_and_back(obj)
    assert_array_equal(obj.predict(z), obj2.predict(z))
    return obj, obj2


def _fit_score_test(
    obj: BaseEstimator,
    x_train: np.ndarray,
    x_test: np.ndarray,
    y_train: Optional[np.ndarray] = None,
    y_test: Optional[np.ndarray] = None,
) -> Tuple[BaseEstimator, BaseEstimator]:
    obj = obj.fit(x_train, y_train)
    obj2 = _to_json_and_back(obj)
    assert_array_equal(obj.score(x_test, y_test), obj2.score(x_test, y_test))
    return obj, obj2


def _fit_transform_x_test(
    obj: BaseEstimator,
    x_train: np.ndarray,
    x_test: np.ndarray,
) -> Tuple[BaseEstimator, BaseEstimator]:
    obj = obj.fit(x_train)
    obj2 = _to_json_and_back(obj)
    assert_array_equal(obj.transform(x_test), obj2.transform(x_test))
    return obj, obj2


def _fit_transform_x_y_test(
    obj: BaseEstimator,
    x_train: np.ndarray,
    x_test: np.ndarray,
    y_train: np.ndarray,
    y_test: np.ndarray,
) -> Tuple[BaseEstimator, BaseEstimator]:
    obj = obj.fit(x_train, y_train)
    obj2 = _to_json_and_back(obj)
    x_tr1, y_tr1 = obj.transform(x_test, y_test)
    x_tr2, y_tr2 = obj2.transform(x_test, y_test)
    assert_array_equal(x_tr1, x_tr2)
    assert_array_equal(y_tr1, y_tr2)
    return obj, obj2


def _test_cross_decomposition_estimator(
    obj: BaseEstimator,
) -> Tuple[BaseEstimator, BaseEstimator]:
    x = [[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [2.0, 2.0, 2.0], [2.0, 5.0, 4.0]]
    y = [[0.1, -0.2], [0.9, 1.1], [6.2, 5.9], [11.9, 12.3]]
    obj = obj.fit(x, y)
    obj2 = _to_json_and_back(obj)
    x1, y1 = obj.transform(x, y)
    x2, y2 = obj2.transform(x, y)
    assert_array_equal(x1, x2)
    assert_array_equal(y1, y2)
    return obj, obj2


def _test_covariance_estimator(
    obj: BaseEstimator,
) -> Tuple[BaseEstimator, BaseEstimator]:
    """Unit tests for covariance estimators"""
    np.random.seed(0)
    true_cov = np.array(
        [
            [0.8, 0.0, 0.2, 0.0],
            [0.0, 0.4, 0.0, 0.0],
            [0.2, 0.0, 0.3, 0.1],
            [0.0, 0.0, 0.1, 0.7],
        ]
    )
    kw = {"mean": [0, 0, 0, 0], "cov": true_cov, "size": 200}
    x = np.random.multivariate_normal(**kw)
    x_test = np.random.multivariate_normal(**kw)
    obj = obj.fit(x)
    obj2 = _to_json_and_back(obj)
    assert_array_equal(obj.score(x_test), obj2.score(x_test))
    assert_array_equal(obj.mahalanobis(x_test), obj2.mahalanobis(x_test))
    return obj, obj2


def test_affinitypropagation():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.AffinityPropagation.html
    """
    x = np.array([[1, 2], [1, 4], [1, 0], [4, 2], [4, 4], [4, 0]])
    y = [[0, 0], [4, 4]]
    _fit_predict_x_z_test(AffinityPropagation(random_state=0), x, y)


def test_agglomerativeclustering():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.AgglomerativeClustering.html
    """
    x = np.array([[1, 2], [1, 4], [1, 0], [4, 2], [4, 4], [4, 0]])
    _fit_labels_test(AgglomerativeClustering(), x)


def test_birch():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.Birch.html
    """
    # x = [[0, 1], [0.3, 1], [-0.3, 1], [0, -1], [0.3, -1], [-0.3, -1]]
    # e = Birch(n_clusters=None)
    # e.fit(x)
    # e2 = _to_json_and_back(e)
    # assert_array_equal(e.predict(x), e2.predict(x))


def test_dbscan():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    """
    x = np.array([[1, 2], [2, 2], [2, 3], [8, 7], [8, 8], [25, 80]])
    _fit_labels_test(DBSCAN(eps=3, min_samples=2), x)


def test_optics():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.OPTICS.html
    """
    x = np.array([[1, 2], [2, 5], [3, 6], [8, 7], [8, 8], [7, 3]])
    _fit_labels_test(OPTICS(min_samples=2), x)


def test_kmeans():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html
    """
    x = np.array([[1, 2], [1, 4], [1, 0], [10, 2], [10, 4], [10, 0]])
    y = [[0, 0], [12, 3]]
    _fit_predict_x_z_test(
        KMeans(n_clusters=2, random_state=0, n_init="auto"), x, y
    )


def test_bisectingkmeans():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.BisectingKMeans.html
    """
    # x = np.array(
    #     [
    #         [1, 2],
    #         [1, 4],
    #         [1, 0],
    #         [10, 2],
    #         [10, 4],
    #         [10, 0],
    #         [10, 6],
    #         [10, 8],
    #         [10, 10],
    #     ]
    # )
    # y = [[0, 0], [12, 3]]
    # e = BisectingKMeans(n_clusters=3, random_state=0).fit(x)
    # e2 = _to_json_and_back(e)
    # assert_array_equal(e.predict(y), e2.predict(y))


def test_minibatchkmeans():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.MiniBatchKMeans.html
    """
    x = np.array(
        [
            [1, 2],
            [1, 4],
            [1, 0],
            [4, 2],
            [4, 0],
            [4, 4],
            [4, 5],
            [0, 1],
            [2, 2],
            [3, 2],
            [5, 5],
            [1, -1],
        ]
    )
    z = [[0, 0], [4, 4]]
    e = MiniBatchKMeans(
        n_clusters=2, random_state=0, batch_size=6, max_iter=10, n_init="auto"
    )
    _fit_predict_x_z_test(e, x, z)


def test_meanshift():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.MeanShift.html
    """
    x = np.array([[1, 1], [2, 1], [1, 0], [4, 7], [3, 5], [3, 6]])
    y = [[0, 0], [5, 5]]
    _fit_predict_x_z_test(MeanShift(bandwidth=2), x, y)


def test_spectralclustering():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.SpectralClustering.html
    """
    x = np.array([[1, 1], [2, 1], [1, 0], [4, 7], [3, 5], [3, 6]])
    e = SpectralClustering(
        n_clusters=2, assign_labels="discretize", random_state=0
    )
    _fit_labels_test(e, x)


def test_spectralbiclustering():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.SpectralBiclustering.html
    """
    x = np.array([[1, 1], [2, 1], [1, 0], [4, 7], [3, 5], [3, 6]])
    e = SpectralBiclustering(n_clusters=2, random_state=0).fit(x)
    e2 = _to_json_and_back(e)
    assert_array_equal(e.row_labels_, e2.row_labels_)
    assert_array_equal(e.column_labels_, e2.column_labels_)


def test_spectralcoclustering():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.SpectralCoclustering.html
    """
    x = np.array([[1, 1], [2, 1], [1, 0], [4, 7], [3, 5], [3, 6]])
    e = SpectralCoclustering(n_clusters=2, random_state=0).fit(x)
    e2 = _to_json_and_back(e)
    assert_array_equal(e.row_labels_, e2.row_labels_)
    assert_array_equal(e.column_labels_, e2.column_labels_)


def test_ellipticenvelope():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.EllipticEnvelope.html
    """
    np.random.seed(0)
    true_cov = np.array([[0.8, 0.3], [0.3, 0.4]])
    x = np.random.multivariate_normal(mean=[0, 0], cov=true_cov, size=500)
    y = [[0, 0], [3, 3]]
    _fit_predict_x_z_test(EllipticEnvelope(random_state=0), x, y)


def test_empiricalcovariance():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.EmpiricalCovariance.html
    """
    _test_covariance_estimator(EmpiricalCovariance())


def test_graphicallasso():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.GraphicalLasso.html
    """
    _test_covariance_estimator(GraphicalLasso())


def test_ledoitwolf():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.LedoitWolf.html
    """
    _test_covariance_estimator(LedoitWolf())


def test_mincovdet():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.MinCovDet.html
    """
    _test_covariance_estimator(MinCovDet(random_state=0))


def test_oas():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.OAS.html
    """
    _test_covariance_estimator(OAS())


def test_shrunkcovariance():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.ShrunkCovariance.html
    """
    _test_covariance_estimator(ShrunkCovariance())


def test_plscanonical():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.PLSCanonical.html
    """
    _test_cross_decomposition_estimator(PLSCanonical(n_components=2))


def test_plsregression():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.PLSRegression.html
    """
    _test_cross_decomposition_estimator(PLSRegression(n_components=2))


def test_plssvd():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.PLSSVD.html
    """
    _test_cross_decomposition_estimator(PLSSVD(n_components=2))


def test_cca():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.CCA.html
    """
    _test_cross_decomposition_estimator(CCA(n_components=1))


def test_dictionarylearning():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.DictionaryLearning.html
    """
    x, _, _ = make_sparse_coded_signal(
        n_samples=100,
        n_components=15,
        n_features=20,
        n_nonzero_coefs=10,
        random_state=42,
        data_transposed=False,
    )
    e = DictionaryLearning(
        n_components=15,
        transform_algorithm="lasso_lars",
        transform_alpha=0.1,
        random_state=42,
        max_iter=10,
    )
    _fit_transform_x_test(e, x, x)


def test_fastica():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.FastICA.html
    """
    # x, _ = load_digits(return_X_y=True)
    # e = FastICA(n_components=7, random_state=0, whiten="unit-variance")
    # _fit_transform_x_test(e, x, x)


def test_incrementalpca():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.IncrementalPCA.html
    """
    x, _ = load_digits(return_X_y=True)
    y = sparse.csr_matrix(x)
    e = IncrementalPCA(n_components=7, batch_size=200)
    e.partial_fit(x[:100, :])
    e2 = _to_json_and_back(e)
    assert_array_equal(e.fit_transform(y), e2.fit_transform(y))


def test_kernelpca():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.KernelPCA.html
    """
    # x, _ = load_digits(return_X_y=True)
    # e = KernelPCA(n_components=7, kernel="linear")
    # _fit_transform_x_test(e, x, x)


def test_minibatchdictionarylearning():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.MiniBatchDictionaryLearning.html
    """


def test_minibatchnmf():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.MiniBatchNMF.html
    """
    # x, _, _ = make_sparse_coded_signal(
    #     n_samples=100,
    #     n_components=15,
    #     n_features=20,
    #     n_nonzero_coefs=10,
    #     random_state=42,
    #     data_transposed=False,
    # )
    # e = MiniBatchDictionaryLearning(
    #     n_components=15,
    #     batch_size=3,
    #     transform_algorithm="lasso_lars",
    #     transform_alpha=0.1,
    #     random_state=42,
    # )
    # _fit_transform_x_test(e, x, x)


def test_minibatchsparsepca():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.MiniBatchSparsePCA.html
    """
    x, _ = make_friedman1(n_samples=200, n_features=30, random_state=0)
    e = MiniBatchSparsePCA(
        n_components=5, batch_size=50, max_iter=10, random_state=0
    )
    e.fit(x)
    _fit_transform_x_test(e, x, x)


def test_nmf():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.NMF.html
    """
    x = np.array([[1, 1], [2, 1], [3, 1.2], [4, 1], [5, 0.8], [6, 1]])
    e = NMF(n_components=2, init="random", random_state=0)
    _fit_transform_x_test(e, x, x)


def test_pca():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
    """
    x = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    e = PCA(n_components=2)
    e, e2 = _fit_transform_x_test(e, x, x)
    assert_array_equal(e.score(x), e2.score(x))


def test_sparsepca():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.SparsePCA.html
    """
    x, _ = make_friedman1(n_samples=200, n_features=30, random_state=0)
    _fit_transform_x_test(SparsePCA(n_components=5, random_state=0), x, x)


def test_factoranalysis():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.FactorAnalysis.html
    """
    x, _ = load_digits(return_X_y=True)
    x = x[:10]
    e = FactorAnalysis(n_components=7, random_state=0, max_iter=10)
    _fit_transform_x_test(e, x, x)


def test_truncatedsvd():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html
    """
    np.random.seed(0)
    x_dense = np.random.rand(100, 100)
    x_dense[:, 2 * np.arange(50)] = 0
    x = sparse.csr_matrix(x_dense)
    e = TruncatedSVD(n_components=5, n_iter=7, random_state=42)
    _fit_transform_x_test(e, x, x)


def test_latentdirichletallocation():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.LatentDirichletAllocation.html
    """
    # x, _ = make_multilabel_classification(random_state=0)
    # e = LatentDirichletAllocation(n_components=5, random_state=0)
    # _fit_transform_x_test(e, x, x[-2:])


def test_randomforestclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    """
    # x, y = make_classification(
    #     n_samples=1000,
    #     n_features=4,
    #     n_informative=2,
    #     n_redundant=0,
    #     random_state=0,
    #     shuffle=False,
    # )
    # z = [[0, 0, 0, 0]]
    # e = RandomForestClassifier(max_depth=2, random_state=0)
    # e.fit(x, y)
    # e2 = _to_json_and_back(e)
    # assert_array_equal(e.predict(z), e2.predict(z))


def test_randomforestregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html
    """
    # x, y = make_classification(
    #     n_samples=1000,
    #     n_features=4,
    #     n_informative=2,
    #     n_redundant=0,
    #     random_state=0,
    #     shuffle=False,
    # )
    # z = [[0, 0, 0, 0]]
    # e = RandomForestRegressor(max_depth=2, random_state=0)
    # e.fit(x, y)
    # e2 = _to_json_and_back(e)
    # assert_array_equal(e.predict(z), e2.predict(z))


def test_randomtreesembedding():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomTreesEmbedding.html
    """
    # x = [[0,0], [1,0], [0,1], [-1,0], [0,-1]]
    # e = RandomTreesEmbedding(n_estimators=5, random_state=0, max_depth=1)
    # _fit_transform_x_test(e, x, x)


def test_extratreesclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    """
    # x, y = make_classification(n_features=4, random_state=0)
    # z = [[0, 0, 0, 0]]
    # e = ExtraTreesClassifier(n_estimators=100, random_state=0)
    # _fit_predict_x_y_z_test(e, x, y, z)


def test_extratreesregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesRegressor.html
    """
    # x, y = load_diabetes(return_X_y=True)
    # x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=0)
    # e = ExtraTreesRegressor(n_estimators=100, random_state=0)
    # _fit_score_test(e, x_train, x_test, y_train, y_test)


def test_baggingclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.BaggingClassifier.html
    """
    # x, y = make_classification(
    #     n_samples=100,
    #     n_features=4,
    #     n_informative=2,
    #     n_redundant=0,
    #     random_state=0,
    #     shuffle=False,
    # )
    # z = [[0, 0, 0, 0]]
    # e = BaggingClassifier(estimator=SVC(), n_estimators=10, random_state=0)
    # _fit_predict_x_y_z_test(e, x, y, z)


def test_baggingregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.BaggingRegressor.html
    """
    # x, y = make_classification(
    #     n_samples=100,
    #     n_features=4,
    #     n_informative=2,
    #     n_redundant=0,
    #     random_state=0,
    #     shuffle=False,
    # )
    # z = [[0, 0, 0, 0]]
    # e = BaggingClassifier(estimator=SVR(), n_estimators=10, random_state=0)
    # _fit_predict_x_y_z_test(e, x, y, z)


def test_isolationforest():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html
    """
    # x = [[-1.1], [0.3], [0.5], [100]]
    # _fit_predict_x_y_test(IsolationForest(random_state=0), x, x)


def test_gradientboostingclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.GradientBoostingClassifier.html
    """
    # x, y = make_hastie_10_2(random_state=0)
    # x_train, x_test = x[:2000], x[2000:]
    # y_train, y_test = y[:2000], y[2000:]
    # e = GradientBoostingClassifier(
    #     n_estimators=100, learning_rate=1.0, max_depth=1, random_state=0
    # )
    # _fit_score_test(e, x_train, x_test, y_train, y_test)


def test_gradientboostingregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.GradientBoostingRegressor.html
    """
    # x, y = make_regression(random_state=0)
    # x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=0)
    # e = GradientBoostingRegressor(random_state=0)
    # e, e2 = _fit_predict_x_y_z_test(e, x_train, y_train, x_test[1:2])
    # assert_array_equal(e.score(x_test, y_test), e2.score(x_test, y_test))


def test_adaboostclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostClassifier.html
    """
    # x, y = make_classification(
    #     n_samples=1000,
    #     n_features=4,
    #     n_informative=2,
    #     n_redundant=0,
    #     random_state=0,
    #     shuffle=False,
    # )
    # z = [[0, 0, 0, 0]]
    # e = AdaBoostClassifier(n_estimators=100, random_state=0)
    # e, e2 = _fit_predict_x_y_z_test(e, x, y, z)
    # assert_array_equal(e.score(x, y), e2.score(x, y))


def test_adaboostregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostRegressor.html
    """
    # x, y = make_classification(
    #     n_samples=1000,
    #     n_features=4,
    #     n_informative=2,
    #     n_redundant=0,
    #     random_state=0,
    #     shuffle=False,
    # )
    # z = [[0, 0, 0, 0]]
    # e = AdaBoostRegressor(random_state=0, n_estimators=100)
    # e, e2 = _fit_predict_x_y_z_test(e, x, y, z)
    # assert_array_equal(e.score(x, y), e2.score(x, y))


def test_histgradientboostingclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.HistGradientBoostingClassifier.html
    """
    # x, y = load_iris(return_X_y=True)
    # e = HistGradientBoostingClassifier(max_iter=10)
    # _fit_score_test(e, x, x, y, y)


def test_histgradientboostingregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.HistGradientBoostingRegressor.html
    """
    # x, y = load_diabetes(return_X_y=True)
    # _fit_score_test(HistGradientBoostingRegressor(), x, x, y, y)


def test_variancethreshold():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.VarianceThreshold.html
    """
    x = [[0, 2, 0, 3], [0, 1, 4, 3], [0, 1, 1, 3]]
    _fit_transform_x_test(VarianceThreshold(), x, x)


def test_gaussianprocessregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.GaussianProcessRegressor.html
    """
    # x, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    # kernel = DotProduct() + WhiteKernel()
    # e = GaussianProcessRegressor(kernel=kernel, random_state=0)
    # e, e2 = _fit_score_test(e, x, x, y, y)
    # assert_array_equal(
    #     e.predict(x[:2, :], return_std=True),
    #     e2.predict(x[:2, :], return_std=True),
    # )
    # assert_array_equal(
    #     e.predict(x[:2, :], return_cov=True),
    #     e2.predict(x[:2, :], return_cov=True),
    # )


def test_gaussianprocessclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.GaussianProcessClassifier.html
    """
    # x, y = load_iris(return_X_y=True)
    # kernel = 1.0 * RBF(1.0)
    # e = GaussianProcessClassifier(
    #     kernel=kernel, random_state=0, max_iter_predict=10
    # )
    # e, e2 = _fit_score_test(e, x, x, y, y)
    # assert_array_equal(
    #     e.predict_proba(x[:2, :]),
    #     e2.predict_proba(x[:2, :]),
    # )


def test_isotonicregression():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.isotonic.IsotonicRegression.html
    """
    # x, y = make_regression(n_samples=10, n_features=1, random_state=41)
    # z = [.1, .2]
    # _fit_predict_x_y_z_test(IsotonicRegression(), x, y, z)


def test_ardregression():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ARDRegression.html
    """
    x, y, z = [[0, 0], [1, 1], [2, 2]], [0, 1, 2], [[1, 1]]
    _fit_predict_x_y_z_test(ARDRegression(n_iter=10), x, y, z)


def test_bayesianridge():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html
    """
    x, y, z = [[0, 0], [1, 1], [2, 2]], [0, 1, 2], [[1, 1]]
    _fit_predict_x_y_z_test(BayesianRidge(), x, y, z)


def test_elasticnet():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html
    """
    x, y = make_regression(n_features=2, random_state=0)
    z = [[0, 0]]
    _fit_predict_x_y_z_test(ElasticNet(random_state=0, max_iter=10), x, y, z)


def test_huberregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.HuberRegressor.html
    """
    rng = np.random.RandomState(0)
    x, y, _ = make_regression(
        n_samples=200, n_features=2, noise=4.0, coef=True, random_state=0
    )
    x[:4] = rng.uniform(10, 20, (4, 2))
    y[:4] = rng.uniform(10, 20, 4)
    e = HuberRegressor()
    e, e2 = _fit_score_test(e, x, x, y, y)
    assert_array_equal(
        e.predict(x[:1, :]),
        e2.predict(x[:1, :]),
    )


def test_lars():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lars.html
    """
    x, y = [[-1, 1], [0, 0], [1, 1]], [-1.1111, 0, -1.1111]
    _fit_predict_x_y_z_test(Lars(n_nonzero_coefs=1), x, y, x)


def test_lasso():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html
    """
    x, y = [[-1, 1], [0, 0], [1, 1]], [-1.1111, 0, -1.1111]
    _fit_predict_x_y_z_test(Lasso(alpha=0), x, y, x)


def test_lassolars():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLars.html
    """
    x, y = [[-1, 1], [0, 0], [1, 1]], [-1.1111, 0, -1.1111]
    _fit_predict_x_y_z_test(LassoLars(alpha=0.01), x, y, x)


def test_lassolarsic():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLarsIC.html
    """
    x = [[-2, 2], [-1, 1], [0, 0], [1, 1], [2, 2]]
    y = [-2.2222, -1.1111, 0, -1.1111, -2.2222]
    _fit_predict_x_y_z_test(LassoLarsIC(criterion="bic"), x, y, x)


def test_linearregression():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html
    """
    x = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
    y = np.dot(x, np.array([1, 2])) + 3
    z = np.array([[3, 5]])
    e, e2 = _fit_score_test(LinearRegression(), x, x, y, y)
    assert_array_equal(e.predict(z), e2.predict(z))


def test_logisticregression():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html
    """
    x, y = load_iris(return_X_y=True)
    e = LogisticRegression(random_state=0, max_iter=100).fit(x, y)
    e2 = _to_json_and_back(e)
    assert_array_equal(e.score(x, y), e2.score(x, y))


def test_multitaskelasticnet():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskElasticNet.html
    """
    x, y = [[0, 0], [1, 1], [2, 2]], [[0, 0], [1, 1], [2, 2]]
    _fit_score_test(MultiTaskElasticNet(alpha=0.1), x, x, y, y)


def test_multitasklasso():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.MultiTaskLasso.html
    """
    x, y = [[0, 0], [1, 1], [2, 2]], [[0, 0], [1, 1], [2, 2]]
    _fit_score_test(MultiTaskLasso(alpha=0.1), x, x, y, y)


def test_orthogonalmatchingpursuit():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.OrthogonalMatchingPursuit.html
    """
    x, y = make_regression(noise=4, random_state=0)
    e, e2 = _fit_score_test(OrthogonalMatchingPursuit(), x, x, y, y)
    assert_array_equal(e.predict(x[:1, :]), e2.predict(x[:1, :]))


def test_passiveaggressiveclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.PassiveAggressiveClassifier.html
    """
    # x, y = make_regression(noise=4, random_state=0)
    # e = PassiveAggressiveClassifier(max_iter=10, random_state=0)
    # e, e2 = _fit_score_test(e, x, x, y, y)
    # assert_array_equal(e.predict(x), e2.predict(x))


def test_passiveaggressiveregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.PassiveAggressiveRegressor.html
    """
    x, y = make_regression(noise=4, random_state=0)
    e = PassiveAggressiveRegressor(max_iter=10, random_state=0)
    e, e2 = _fit_score_test(e, x, x, y, y)
    assert_array_equal(e.predict(x), e2.predict(x))


def test_perceptron():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Perceptron.html
    """
    # x, y = load_digits(return_X_y=True)
    # _fit_score_test(Perceptron(tol=1e-3, random_state=0), x, x, y, y)


def test_quantileregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.QuantileRegressor.html
    """
    n_samples, n_features = 10, 2
    rng = np.random.RandomState(0)
    x, y = rng.randn(n_samples, n_features), rng.randn(n_samples)
    e = QuantileRegressor(quantile=0.8, solver="interior-point")
    _fit_score_test(e, x, x, y, y)


def test_ridge():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html
    """
    n_samples, n_features = 10, 5
    rng = np.random.RandomState(0)
    x, y = rng.randn(n_samples, n_features), rng.randn(n_samples)
    _fit_score_test(Ridge(alpha=1.0), x, x, y, y)


def test_ridgeclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RidgeClassifier.html
    """
    x, y = load_breast_cancer(return_X_y=True)
    _fit_score_test(RidgeClassifier(), x, x, y, y)


def test_sgdclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDClassifier.html
    """
    # x = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
    # y = np.array([1, 1, 2, 2])
    # x = StandardScaler().fit_transform(x)
    # _fit_score_test(SGDClassifier(max_iter=10, tol=1e-3), x, x, y, y)


def test_sgdregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDRegressor.html
    """
    n_samples, n_features = 10, 5
    rng = np.random.RandomState(0)
    x, y = rng.randn(n_samples, n_features), rng.randn(n_samples)
    x = StandardScaler().fit_transform(x)
    _fit_score_test(SGDRegressor(max_iter=10, tol=1e-3), x, x, y, y)


def test_sgdoneclasssvm():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDOneClassSVM.html
    """
    # x = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
    # _fit_predict_x_z_test(SGDOneClassSVM(random_state=42), x, x)


def test_theilsenregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.TheilSenRegressor.html
    """
    x, y = make_regression(
        n_samples=200, n_features=2, noise=4.0, random_state=0
    )
    _fit_score_test(TheilSenRegressor(random_state=0), x, x, y, y)


def test_ransacregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RANSACRegressor.html
    """
    x, y = make_regression(
        n_samples=200, n_features=2, noise=4.0, random_state=0
    )
    _fit_score_test(RANSACRegressor(random_state=0), x, x, y, y)


def test_poissonregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.PoissonRegressor.html
    """
    # x, y = [[1, 2], [2, 3], [3, 4], [4, 3]], [12, 17, 22, 21]
    # _fit_score_test(PoissonRegressor(), x, x, y, y)


def test_gammaregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.GammaRegressor.html
    """
    # x, y = [[1, 2], [2, 3], [3, 4], [4, 3]], [19, 26, 33, 30]
    # _fit_score_test(GammaRegressor(), x, x, y, y)


def test_tweedieregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.TweedieRegressor.html
    """
    # x, y = [[1, 2], [2, 3], [3, 4], [4, 3]], [2, 3.5, 5, 5.5]
    # _fit_score_test(TweedieRegressor(), x, x, y, y)


def test_locallylinearembedding():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.LocallyLinearEmbedding.html
    """
    x, _ = load_digits(return_X_y=True)
    e = LocallyLinearEmbedding(n_components=2)
    xe = e.fit_transform(x[:100])
    e2 = _to_json_and_back(e)
    assert_array_equal(xe, e2.embedding_)


def test_isomap():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.Isomap.html
    """
    x, _ = load_digits(return_X_y=True)
    e = Isomap(n_components=2)
    xe = e.fit_transform(x[:100])
    e2 = _to_json_and_back(e)
    assert_array_equal(xe, e2.embedding_)


def test_mds():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.MDS.html
    """
    x, _ = load_digits(return_X_y=True)
    e = MDS(n_components=2, normalized_stress="auto")
    xe = e.fit_transform(x[:100])
    e2 = _to_json_and_back(e)
    assert_array_equal(xe, e2.embedding_)


def test_spectralembedding():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.SpectralEmbedding.html
    """
    x, _ = load_digits(return_X_y=True)
    e = SpectralEmbedding(n_components=2)
    xe = e.fit_transform(x[:100])
    e2 = _to_json_and_back(e)
    assert_array_equal(xe, e2.embedding_)


def test_tsne():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
    """
    x = np.array([[0, 0, 0], [0, 1, 1], [1, 0, 1], [1, 1, 1]])
    e = TSNE(n_components=2, perplexity=3, n_iter=250)
    xe = e.fit_transform(x)
    e2 = _to_json_and_back(e)
    assert_array_equal(xe, e2.embedding_)


def test_gaussianmixture():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.mixture.GaussianMixture.html
    """
    x = np.array([[1, 2], [1, 4], [1, 0], [10, 2], [10, 4], [10, 0]])
    gm = GaussianMixture(n_components=2, random_state=0).fit(x)
    gm2 = _to_json_and_back(gm)
    y = [[0, 0], [12, 3]]
    assert_array_equal(gm.predict(y), gm2.predict(y))
    assert_array_equal(gm.score(x), gm2.score(x))


def test_bayesiangaussianmixture():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.mixture.BayesianGaussianMixture.html
    """
    x = np.array([[1, 2], [1, 4], [1, 0], [10, 2], [10, 4], [10, 0]])
    gm = BayesianGaussianMixture(n_components=2, random_state=0).fit(x)
    gm2 = _to_json_and_back(gm)
    y = [[0, 0], [12, 3]]
    assert_array_equal(gm.predict(y), gm2.predict(y))
    assert_array_equal(gm.score(x), gm2.score(x))


def test_bernoullinb():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.BernoulliNB.html
    """
    rng = np.random.RandomState(1)
    x, y = rng.randint(5, size=(6, 100)), np.array([1, 2, 3, 4, 4, 5])
    _fit_predict_x_y_z_test(BernoulliNB(force_alpha=True), x, y, x[2:3])


def test_gaussiannb():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.GaussianNB.html
    """
    x = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    y = np.array([1, 1, 1, 2, 2, 2])
    z = [[-0.8, -1]]
    _fit_predict_x_y_z_test(GaussianNB(), x, y, z)


def test_multinomialnb():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.MultinomialNB.html
    """
    rng = np.random.RandomState(1)
    x, y = rng.randint(5, size=(6, 100)), np.array([1, 2, 3, 4, 5, 6])
    _fit_predict_x_y_z_test(MultinomialNB(force_alpha=True), x, y, x[2:3])


def test_complementnb():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.ComplementNB.html
    """
    rng = np.random.RandomState(1)
    x, y = rng.randint(5, size=(6, 100)), np.array([1, 2, 3, 4, 5, 6])
    _fit_predict_x_y_z_test(ComplementNB(force_alpha=True), x, y, x[2:3])


def test_categoricalnb():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.CategoricalNB.html
    """
    rng = np.random.RandomState(1)
    x, y = rng.randint(5, size=(6, 100)), np.array([1, 2, 3, 4, 5, 6])
    _fit_predict_x_y_z_test(CategoricalNB(force_alpha=True), x, y, x[2:3])


def test_kneighborsclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html
    """
    # x, y, z = [[0], [1], [2], [3]], [0, 0, 1, 1], [[1.1]]
    # _fit_predict_x_y_z_test(KNeighborsClassifier(n_neighbors=3), x, y, z)


def test_kneighborsregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsRegressor.html
    """
    # x, y, z = [[0], [1], [2], [3]], [0, 0, 1, 1], [[1.1]]
    # _fit_predict_x_y_z_test(KNeighborsRegressor(n_neighbors=3), x, y, z)


def test_kneighborstransformer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsTransformer.html
    """
    # x, _ = load_wine(return_X_y=True)
    # e = KNeighborsTransformer(n_neighbors=5, mode="distance")
    # _fit_transform_x_test(e, x, x)


def test_nearestcentroid():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.NearestCentroid.html
    """
    x = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    y = np.array([1, 1, 1, 2, 2, 2])
    z = [[-0.8, -1]]
    _fit_predict_x_y_z_test(NearestCentroid(), x, y, z)


def test_nearestneighbors():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.NearestNeighbors.html
    """
    x, z = [[0, 0, 2], [1, 0, 0], [0, 0, 1]], [[0, 0, 1.3]]
    e = NearestNeighbors(n_neighbors=2, radius=0.4)
    e.fit(x)
    e2 = _to_json_and_back(e)
    assert_array_equal(
        e.kneighbors(z, 2, return_distance=False),
        e2.kneighbors(z, 2, return_distance=False),
    )


def test_radiusneighborsclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.RadiusNeighborsClassifier.html
    """
    # x, y, z = [[0], [1], [2], [3]], [0, 0, 1, 1], [[1.5]]
    # _fit_predict_x_y_z_test(RadiusNeighborsClassifier(radius=1.0), x, y, z)


def test_radiusneighborsregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.RadiusNeighborsRegressor.html
    """
    # x, y, z = [[0], [1], [2], [3]], [0, 0, 1, 1], [[1.5]]
    # _fit_predict_x_y_z_test(RadiusNeighborsRegressor(radius=1.0), x, y, z)


def test_radiusneighborstransformer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.RadiusNeighborsTransformer.html
    """
    # x, _ = load_wine(return_X_y=True)
    # e = RadiusNeighborsTransformer(radius=42.0, mode="distance")
    # _fit_transform_x_test(e, x, x)


def test_kerneldensity():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KernelDensity.html
    """
    # rng = np.random.RandomState(42)
    # x = rng.random_sample((100, 3))
    # e = KernelDensity(kernel="gaussian", bandwidth=0.5).fit(x)
    # e2 = _to_json_and_back(e)
    # assert_array_equal(e.score_samples(x[:3]), e2.score_samples(x[:3]))


def test_localoutlierfactor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.LocalOutlierFactor.html
    """
    # x = [[-1.1], [0.2], [101.1], [0.3]]
    # e = LocalOutlierFactor(n_neighbors=2, novelty=True)
    # _fit_predict_x_z_test(e, x, x)


def test_neighborhoodcomponentsanalysis():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.NeighborhoodComponentsAnalysis.html
    """
    # x, y = load_iris(return_X_y=True)
    # e = NeighborhoodComponentsAnalysis(random_state=42)
    # _fit_transform_x_y_test(e, x, x, y, y)


def test_bernoullirbm():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.BernoulliRBM.html
    """
    x = np.array([[0, 0, 0], [0, 1, 1], [1, 0, 1], [1, 1, 1]])
    _fit_transform_x_test(BernoulliRBM(n_components=2), x, x)


def test_mlpclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPClassifier.html
    """
    # x, y = make_classification(n_samples=100, random_state=1)
    # x_train, x_test, y_train, y_test = train_test_split(
    #     x, y, stratify=y, random_state=1
    # )
    # e = MLPClassifier(random_state=1, max_iter=10)
    # _fit_score_test(e, x_train, x_test, y_train, y_test)


def test_mlpregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html
    """
    # x, y = make_classification(n_samples=100, random_state=1)
    # x_train, x_test, y_train, y_test = train_test_split(
    #     x, y, stratify=y, random_state=1
    # )
    # e = MLPRegressor(random_state=1, max_iter=10)
    # _fit_score_test(e, x_train, x_test, y_train, y_test)


def test_binarizer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Binarizer.html
    """
    x = [[1.0, -1.0, 2.0], [2.0, 0.0, 0.0], [0.0, 1.0, -1.0]]
    _fit_transform_x_test(Binarizer(), x, x)


def test_kbinsdiscretizer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.KBinsDiscretizer.html
    """
    # x = [[-2, 1, -4, -1], [-1, 2, -3, -0.5], [0, 3, -2, 0.5], [1, 4, -1, 2]]
    # e = KBinsDiscretizer(n_bins=3, encode="ordinal", strategy="uniform")
    # _fit_transform_x_test(e, x, x)


def test_kernelcenterer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.KernelCenterer.html
    """
    x = [[1.0, -2.0, 2.0], [-2.0, 1.0, 3.0], [4.0, 1.0, -2.0]]
    k = pairwise_kernels(x, metric="linear")
    _fit_transform_x_test(KernelCenterer(), k, k)


def test_labelbinarizer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelBinarizer.html
    """
    x, z = [1, 2, 6, 4, 2], [1, 6]
    _fit_transform_x_test(LabelBinarizer(), x, z)


def test_labelencoder():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html
    """
    x = [1, 2, 2, 6]
    _fit_transform_x_test(LabelEncoder(), x, x)


def test_multilabelbinarizer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MultiLabelBinarizer.html
    """
    x = [(1, 2), (3,)]
    _fit_transform_x_test(MultiLabelBinarizer(), x, x)


def test_minmaxscaler():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html
    """
    x = [[-1, 2], [-0.5, 6], [0, 10], [1, 18]]
    _fit_transform_x_test(MinMaxScaler(), x, x)


def test_maxabsscaler():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MaxAbsScaler.html
    """
    x = [[1.0, -1.0, 2.0], [2.0, 0.0, 0.0], [0.0, 1.0, -1.0]]
    _fit_transform_x_test(MaxAbsScaler(), x, x)


def test_quantiletransformer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.QuantileTransformer.html
    """
    rng = np.random.RandomState(0)
    x = np.sort(rng.normal(loc=0.5, scale=0.25, size=(25, 1)), axis=0)
    e = QuantileTransformer(n_quantiles=10, random_state=0)
    _fit_transform_x_test(e, x, x)


def test_normalizer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Normalizer.html
    """
    x = [[4, 1, 2, 2], [1, 3, 9, 3], [5, 7, 5, 1]]
    _fit_transform_x_test(Normalizer(), x, x)


def test_powertransformer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PowerTransformer.html
    """
    # x = [[1, 2], [3, 2], [4, 5]]
    # _fit_transform_x_test(PowerTransformer(), x, x)


def test_robustscaler():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html
    """
    x = [[1.0, -2.0, 2.0], [-2.0, 1.0, 3.0], [4.0, 1.0, -2.0]]
    _fit_transform_x_test(RobustScaler(), x, x)


def test_splinetransformer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.SplineTransformer.html
    """
    # x = np.arange(6).reshape(6, 1)
    # _fit_transform_x_test(SplineTransformer(degree=2, n_knots=3), x, x)


def test_standardscaler():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html
    """
    x = [[0, 0], [0, 0], [1, 1], [1, 1]]
    _fit_transform_x_test(StandardScaler(), x, x)


def test_polynomialfeatures():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PolynomialFeatures.html
    """
    x = np.arange(6).reshape(3, 2)
    _fit_transform_x_test(PolynomialFeatures(2), x, x)


def test_sparserandomprojection():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.random_projection.SparseRandomProjection.html
    """
    # rng = np.random.RandomState(42)
    # x = rng.rand(25, 3000)
    # _fit_transform_x_test(SparseRandomProjection(random_state=rng), x, x)


def test_gaussianrandomprojection():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.random_projection.GaussianRandomProjection.html
    """
    # rng = np.random.RandomState(42)
    # x = rng.rand(25, 3000)
    # _fit_transform_x_test(GaussianRandomProjection(random_state=rng), x, x)


def test_labelpropagation():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.semi_supervised.LabelPropagation.html
    """
    # iris = load_iris()
    # rng = np.random.RandomState(42)
    # random_unlabeled_points = rng.rand(len(iris.target)) < 0.3
    # labels = np.copy(iris.target)
    # labels[random_unlabeled_points] = -1
    # e = LabelPropagation()
    # _fit_score_test(e, iris.data, iris.data, labels, labels)


def test_labelspreading():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.semi_supervised.LabelSpreading.html
    """
    # iris = load_iris()
    # rng = np.random.RandomState(42)
    # random_unlabeled_points = rng.rand(len(iris.target)) < 0.3
    # labels = np.copy(iris.target)
    # labels[random_unlabeled_points] = -1
    # e = LabelSpreading()
    # _fit_score_test(e, iris.data, iris.data, labels, labels)


def test_linearsvc():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVC.html
    """
    x, y = make_classification(n_features=4, random_state=0)
    e = LinearSVC(random_state=0, tol=1e-5)
    _fit_score_test(e, x, x, y, y)


def test_linearsvr():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVR.html
    """
    x, y = make_classification(n_features=4, random_state=0)
    e = LinearSVR(random_state=0, tol=1e-5)
    _fit_score_test(e, x, x, y, y)


def test_nusvc():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.NuSVC.html
    """
    # x = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
    # y = np.array([1, 1, 2, 2])
    # _fit_score_test(NuSVC(), x, x, y, y)


def test_nusvr():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.NuSVR.html
    """
    # x = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
    # y = np.array([1, 1, 2, 2])
    # _fit_score_test(NuSVR(), x, x, y, y)


def test_oneclasssvm():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.OneClassSVM.html
    """
    # x = [[0], [0.44], [0.45], [0.46], [1]]
    # _fit_predict_x_z_test(OneClassSVM(gamma="auto"), x, x)


def test_svc():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html
    """
    # x = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
    # y = np.array([1, 1, 2, 2])
    # z = [[-0.8, -1]]
    # _fit_predict_x_y_z_test(SVC(gamma="auto"), x, y, z)


def test_svr():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html
    """
    # x = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
    # y = np.array([1, 1, 2, 2])
    # z = [[-0.8, -1]]
    # _fit_predict_x_y_z_test(SVR(gamma="auto"), x, y, z)


def test_decisiontreeclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    """
    # iris = load_iris()
    # x, y = iris.data, iris.target
    # e = DecisionTreeClassifier(random_state=0)
    # _fit_score_test(e, x, x, y, y)


def test_decisiontreeregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    """
    # x, y = load_diabetes(return_X_y=True)
    # e = DecisionTreeRegressor(random_state=0)
    # _fit_score_test(e, x, x, y, y)


def test_extratreeclassifier():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.tree.ExtraTreeClassifier.html
    """
    # x, y = load_iris(return_X_y=True)
    # x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=0)
    # e = ExtraTreeClassifier(random_state=0)
    # _fit_score_test(e, x_train, x_test, y_train, y_test)


def test_extratreeregressor():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.tree.ExtraTreeRegressor.html
    """
    # x, y = load_iris(return_X_y=True)
    # x_train, x_test, y_train, y_test = train_test_split(x, y, random_state=0)
    # e = ExtraTreeRegressor(random_state=0)
    # _fit_score_test(e, x_train, x_test, y_train, y_test)


def test_lineardiscriminantanalysis():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.discriminant_analysis.LinearDiscriminantAnalysis.html
    """
    x = [[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]]
    y, z = [1, 1, 1, 2, 2, 2], [[-0.8, -1]]
    _fit_predict_x_y_z_test(LinearDiscriminantAnalysis(), x, y, z)


def test_quadraticdiscriminantanalysis():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis.html
    """
    x = [[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]]
    y, z = [1, 1, 1, 2, 2, 2], [[-0.8, -1]]
    _fit_predict_x_y_z_test(QuadraticDiscriminantAnalysis(), x, y, z)


def test_missingindicator():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.impute.MissingIndicator.html
    """
    # x_train = np.array([[np.nan, 1, 3], [4, 0, np.nan], [8, 1, 0]])
    # x_test = np.array([[5, 1, np.nan], [np.nan, 2, 3], [2, 4, 0]])
    # _fit_transform_x_test(MissingIndicator(), x_train, x_test)


def test_simpleimputer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.impute.SimpleImputer.html
    """
    x_train = np.array([[np.nan, 1, 3], [4, 0, np.nan], [8, 1, 0]])
    x_test = np.array([[5, 1, np.nan], [np.nan, 2, 3], [2, 4, 0]])
    _fit_transform_x_test(SimpleImputer(), x_train, x_test)


def test_knnimputer():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.impute.KNNImputer.html
    """
    x = [[1, 2, np.nan], [3, 4, 3], [np.nan, 6, 5], [8, 8, 7]]
    _fit_transform_x_test(KNNImputer(n_neighbors=2), x, x)


def test_polynomialcountsketch():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.kernel_approximation.PolynomialCountSketch.html
    """
    x = [[0, 0], [1, 1], [1, 0], [0, 1]]
    e = PolynomialCountSketch(degree=3, random_state=1)
    _fit_transform_x_test(e, x, x)


def test_rbfsampler():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.kernel_approximation.RBFSampler.html
    """
    x = [[0, 0], [1, 1], [1, 0], [0, 1]]
    e = RBFSampler(gamma=1, random_state=1)
    _fit_transform_x_test(e, x, x)


def test_skewedchi2sampler():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.kernel_approximation.SkewedChi2Sampler.html
    """
    x = [[0, 0], [1, 1], [1, 0], [0, 1]]
    e = SkewedChi2Sampler(skewedness=0.01, n_components=10, random_state=0)
    _fit_transform_x_test(e, x, x)


def test_additivechi2sampler():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.kernel_approximation.AdditiveChi2Sampler.html
    """
    x, y = load_digits(return_X_y=True)
    e = AdditiveChi2Sampler(sample_steps=2)
    e.fit(x, y)
    e2 = _to_json_and_back(e)
    assert_array_equal(e.transform(x), e2.transform(x))


def test_nystroem():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.kernel_approximation.Nystroem.html
    """
    x, _ = load_digits(n_class=9, return_X_y=True)
    e = Nystroem(gamma=0.2, random_state=1, n_components=300)
    _fit_transform_x_test(e, x, x)


def test_kernelridge():
    """
    https://scikit-learn.org/stable/modules/generated/sklearn.kernel_ridge.KernelRidge.html
    """
    n_samples, n_features = 10, 5
    rng = np.random.RandomState(0)
    x, y = rng.randn(n_samples, n_features), rng.randn(n_samples)
    _fit_score_test(KernelRidge(alpha=1.0), x, x, y, y)



# pylint: disable=missing-function-docstring
"""Decode exclusion tests"""

from collections import deque

import numpy as np
import pandas as pd
import tensorflow as tf
from turbo_broccoli.environment import set_nodecode

from common import from_json, to_json
from test_dataclass import C, D
from test_keras import _build_model
from test_pandas import _assert_equal as assert_equal_pd


def _basic_dict() -> dict:
    return {"a_list": [1, "2", None], "a_str": "abcd", "an_int": 42}


def test_nodecode_nothing():
    set_nodecode("")
    x = _basic_dict()
    assert x == from_json(to_json(x))
    set_nodecode("")


def test_nodecode_bytes():
    set_nodecode("bytes")
    x = {"b": "Hello 🌎".encode("utf8"), **_basic_dict()}
    y = {"b": None, **_basic_dict()}
    assert y == from_json(to_json(x))
    set_nodecode("")


def test_nodecode_dataclass():
    set_nodecode("dataclass.C")
    c = C(a_byte_str=b"", a_list=[], a_str="", an_int=0)
    x = {"dcc": c, "dcd": D(a_dataclass=c, a_float=1.2), **_basic_dict()}
    y = {"dcc": None, "dcd": D(a_dataclass=None, a_float=1.2), **_basic_dict()}
    assert y == from_json(to_json(x))
    set_nodecode("")


def test_nodecode_collections():
    set_nodecode("collections.deque")
    x = {"deq": deque(range(100)), **_basic_dict()}
    y = {"deq": None, **_basic_dict()}
    assert y == from_json(to_json(x))
    set_nodecode("")


def test_nodecode_keras():
    set_nodecode("keras.model")
    x = {"model": _build_model(), **_basic_dict()}
    y = {"model": None, **_basic_dict()}
    assert y == from_json(to_json(x))
    set_nodecode("")


def test_nodecode_numpy():
    set_nodecode("numpy.ndarray")
    x = {"arr": np.zeros(10), **_basic_dict()}
    y = {"arr": None, **_basic_dict()}
    assert y == from_json(to_json(x))
    set_nodecode("")


# The serialization of Series depends on that of DataFrame. So excluding
# pandas.dataframe will make it impossible to deserialize Series.
# def test_nodecode_pandas_df():
#     set_nodecode("pandas.dataframe")
#     s = pd.Series([1, 2, 3])
#     df = pd.DataFrame(
#         {
#             "a": s,
#             "b": pd.Categorical(["X", "Y", "X"])
#         }
#     )
#     x = {
#         "s": s,
#         "df": df,
#     }
#     print(to_json(x))
#     y = from_json(to_json(x))
#     assert_equal_pd(x["s"], y["s"])
#     assert y["df"] is None
#     set_nodecode("")


def test_nodecode_pandas_ser():
    set_nodecode("pandas.series")
    s = pd.Series([1, 2, 3])
    df = pd.DataFrame({"a": s, "b": pd.Categorical(["X", "Y", "X"])})
    x = {
        "s": s,
        "df": df,
    }
    y = from_json(to_json(x))
    assert_equal_pd(x["df"], y["df"])
    assert y["s"] is None
    set_nodecode("")


def test_nodecode_tensorflow():
    set_nodecode("tensorflow.tensor")
    x = {"t": tf.random.uniform((10, 10)), **_basic_dict()}
    y = {"t": None, **_basic_dict()}
    assert y == from_json(to_json(x))
    set_nodecode("")



# pylint: disable=missing-function-docstring
"""Keras (de)serialization test suite"""

from tensorflow import keras
import numpy as np
from numpy.testing import assert_array_equal

from turbo_broccoli.environment import set_keras_format

from common import from_json, to_json


def _assert_model_equal(a, b):
    assert a.get_config() == b.get_config()
    for i, w in enumerate(a.weights):
        assert_array_equal(w, b.weights[i])
    # Not really necessary but why not
    assert_array_equal(a(np.ones((1, 28, 28, 1))), b(np.ones((1, 28, 28, 1))))


def _build_model():
    model = keras.Sequential(
        [
            keras.Input(shape=(28, 28, 1)),
            keras.layers.Conv2D(32, kernel_size=(3, 3), activation="relu"),
            keras.layers.MaxPooling2D(pool_size=(2, 2)),
            keras.layers.Conv2D(64, kernel_size=(3, 3), activation="relu"),
            keras.layers.MaxPooling2D(pool_size=(2, 2)),
            keras.layers.Flatten(),
            keras.layers.Dropout(0.5),
            keras.layers.Dense(10, activation="softmax"),
        ]
    )
    model.compile(
        loss="categorical_crossentropy",
        optimizer="adam",
        metrics=["accuracy"],
    )
    return model


def test_keras_model_json():
    set_keras_format("json")
    x = _build_model()
    _assert_model_equal(x, from_json(to_json(x)))


def test_keras_model_h5():
    set_keras_format("h5")
    x = _build_model()
    _assert_model_equal(x, from_json(to_json(x)))


def test_keras_model_tf():
    set_keras_format("tf")
    x = _build_model()
    _assert_model_equal(x, from_json(to_json(x)))



# pylint: disable=missing-function-docstring
"""Python collections (de)serialization test suite"""

from collections import deque, namedtuple
from typing import NamedTuple

from common import from_json, to_json


def _assert_equal(a: deque, b: deque):
    assert a.maxlen == b.maxlen
    assert a == b


def test_deque_empty():
    x = deque()
    _assert_equal(x, from_json(to_json(x)))
    x = deque([])
    _assert_equal(x, from_json(to_json(x)))


def test_deque_normal():
    x = deque(range(100))
    _assert_equal(x, from_json(to_json(x)))


def test_deque_maxlen_empty():
    x = deque([], maxlen=1)
    _assert_equal(x, from_json(to_json(x)))


def test_deque_maxlen_trunc():
    x = deque(range(100), maxlen=10)
    _assert_equal(x, from_json(to_json(x)))


def test_namedtuple_empty():
    C = namedtuple("C", [])
    x = C()
    assert x == from_json(to_json(x))


def test_namedtuple_normal():
    C = namedtuple("C", ["x", "y"])
    x = C(1, 2)
    assert x == from_json(to_json(x))


def test_namedtuple_default():
    C = namedtuple("C", ["x", "y"], defaults=[0, 1])
    x = C()
    assert x == from_json(to_json(x))


def test_namedtuple_subclass_1():
    # pylint: disable=missing-class-docstring
    class C(namedtuple("C", ["x", "y"])):
        @property
        def z(self):
            return self.x + self.y

    x = C(1, 2)
    assert x == from_json(to_json(x))


def test_namedtuple_subclass_2():
    # pylint: disable=missing-class-docstring
    class C(NamedTuple):
        x: int
        y: int

        @property
        def z(self):
            return self.x + self.y

    x = C(1, 2)
    assert x == from_json(to_json(x))



# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
"""deque (de)serialization test suite"""

from dataclasses import dataclass

from turbo_broccoli.environment import register_dataclass_type

from common import from_json, to_json


@dataclass
class C:
    a_byte_str: bytes
    a_list: list
    a_str: str
    an_int: int


@dataclass
class D:
    a_dataclass: C
    a_float: float


def test_dataclass():
    register_dataclass_type(C)
    x = C(
        a_byte_str="🦐🦐🦐".encode("utf8"),
        a_list=list(range(10)),
        a_str="Hello 🌎",
        an_int="42",
    )
    y = from_json(to_json(x))
    assert isinstance(y, C)
    assert x.__dict__ == y.__dict__


def test_dataclass_recursive():
    register_dataclass_type(C)
    register_dataclass_type(D)
    x = D(
        a_dataclass=C(
            a_byte_str="🦐🦐🦐".encode("utf8"),
            a_list=list(range(10)),
            a_str="Hello 🌎",
            an_int="42",
        ),
        a_float=1.2,
    )
    y = from_json(to_json(x))
    assert isinstance(y, D)
    assert x.a_float == y.a_float
    assert isinstance(x.a_dataclass, C)
    assert x.a_dataclass.__dict__ == y.a_dataclass.__dict__



# pylint: disable=import-outside-toplevel
# pylint: disable=missing-function-docstring
# pylint: disable=unused-wildcard-import
"""
bokeh (de)serialization test suite. This test suite serializes and deserializes
various figures, but does not check for similarity between the original and the
deserialized...

TODO:
    Make every example of the [bokeh
    gallery](https://docs.bokeh.org/en/latest/docs/gallery.html) into a unit
    test. This is probably widely unnecessary though :]
"""

import numpy as np
from bokeh.core.enums import *
from bokeh.layouts import *
from bokeh.models import *
from bokeh.palettes import *
from bokeh.plotting import figure
from bokeh.transform import linear_cmap
from bokeh.util.hex import hexbin
from common import from_json, to_json
from numpy.random import random, standard_normal
from scipy.special import jv
from scipy.stats import gaussian_kde


def test_markers():
    """https://docs.bokeh.org/en/latest/docs/examples/basic/scatters/markers.html"""
    p = figure(title="Bokeh Markers", toolbar_location=None)
    p.grid.grid_line_color = None
    p.background_fill_color = "#eeeeee"
    p.axis.visible = False
    p.y_range.flipped = True
    N = 10
    for i, marker in enumerate(MarkerType):
        x = i % 4
        y = (i // 4) * 4 + 1
        p.scatter(
            random(N) + 2 * x,
            random(N) + y,
            marker=marker,
            size=14,
            line_color="navy",
            fill_color="orange",
            alpha=0.5,
        )
        p.text(
            2 * x + 0.5,
            y + 2.5,
            text=[marker],
            text_color="firebrick",
            text_align="center",
            text_font_size="13px",
        )
    from_json(to_json(p))


def test_color_scatter():
    """https://docs.bokeh.org/en/latest/docs/examples/basic/scatters/color_scatter.html"""
    N = 4000
    x = np.random.random(size=N) * 100
    y = np.random.random(size=N) * 100
    radii = np.random.random(size=N) * 1.5
    colors = np.array(
        [(r, g, 150) for r, g in zip(50 + 2 * x, 30 + 2 * y)], dtype="uint8"
    )
    TOOLS = "hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,tap,save,box_select,poly_select,lasso_select,examine,help"
    p = figure(tools=TOOLS)
    p.scatter(
        x, y, radius=radii, fill_color=colors, fill_alpha=0.6, line_color=None
    )
    from_json(to_json(p))


# def test_elements():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/scatters/elements.html"""
#     from_json(to_json(p))


# def test_image_url():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/scatters/image_url.html"""
#     from_json(to_json(p))


# def test_lorenz():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/lines/lorenz.html"""
#     from_json(to_json(p))


def test_linear_cmap():
    """https://docs.bokeh.org/en/latest/docs/examples/basic/data/linear_cmap.html"""
    x = standard_normal(50000)
    y = standard_normal(50000)
    bins = hexbin(x, y, 0.1)
    p = figure(tools="", match_aspect=True, background_fill_color="#440154")
    p.grid.visible = False
    p.hex_tile(
        q="q",
        r="r",
        size=0.1,
        line_color=None,
        source=bins,
        fill_color=linear_cmap("counts", "Viridis256", 0, max(bins.counts)),
    )
    from_json(to_json(p))


# def test_linear_cmap_colorbar():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/data/linear_cmap_colorbar.html"""
#     from_json(to_json(p))


# def test_color_mappers():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/data/color_mappers.html"""
#     from_json(to_json(p))


# def test_transform_markers():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/data/transform_markers.html"""
#     from_json(to_json(p))


# def test_transform_jitter():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/data/transform_jitter.html"""
#     from_json(to_json(p))


def test_logplot():
    """https://docs.bokeh.org/en/latest/docs/examples/basic/axes/logplot.html"""
    x = np.linspace(0.1, 5, 80)
    p = figure(
        title="log axis example",
        y_axis_type="log",
        x_range=(0, 5),
        y_range=(0.001, 10.0**22),
        background_fill_color="#fafafa",
    )
    p.line(
        x,
        np.sqrt(x),
        legend_label="y=sqrt(x)",
        line_color="tomato",
        line_dash="dashed",
    )
    p.line(x, x, legend_label="y=x")
    p.scatter(x, x, legend_label="y=x")
    p.line(x, x**2, legend_label="y=x**2")
    p.scatter(
        x,
        x**2,
        legend_label="y=x**2",
        fill_color=None,
        line_color="olivedrab",
    )
    p.line(x, 10**x, legend_label="y=10^x", line_color="gold", line_width=2)
    p.line(
        x,
        x**x,
        legend_label="y=x^x",
        line_dash="dotted",
        line_color="indigo",
        line_width=2,
    )
    p.line(
        x,
        10 ** (x**2),
        legend_label="y=10^(x^2)",
        line_color="coral",
        line_dash="dotdash",
        line_width=2,
    )
    p.legend.location = "top_left"
    from_json(to_json(p))


# def test_twin_axes():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/axes/twin_axes.html"""
#     from_json(to_json(p))


def test_fixed_axis():
    """https://docs.bokeh.org/en/latest/docs/examples/basic/axes/fixed_axis.html"""
    x = np.linspace(-6, 6, 500)
    y = 8 * np.sin(x) * np.sinc(x)
    p = figure(
        width=800,
        height=300,
        title="",
        tools="",
        toolbar_location=None,
        match_aspect=True,
    )
    p.line(x, y, color="navy", alpha=0.4, line_width=4)
    p.background_fill_color = "#efefef"
    p.xaxis.fixed_location = 0
    from_json(to_json(p))


# def test_basic():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/basic.html"""
#     from_json(to_json(p))


# def test_colormapped():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/colormapped.html"""
#     from_json(to_json(p))


# def test_intervals():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/intervals.html"""
#     from_json(to_json(p))


# def test_mixed():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/mixed.html"""
#     from_json(to_json(p))


# def test_nested_colormapped():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/nested_colormapped.html"""
#     from_json(to_json(p))


# def test_pandas_groupby_colormapped():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/pandas_groupby_colormapped.html"""
#     from_json(to_json(p))


# def test_pandas_groupby_nested():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/pandas_groupby_nested.html"""
#     from_json(to_json(p))


# def test_stacked():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/stacked.html"""
#     from_json(to_json(p))


# def test_stacked_split():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/stacked_split.html"""
#     from_json(to_json(p))


# def test_nested():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/nested.html"""
#     from_json(to_json(p))


# def test_colors():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/colors.html"""
#     from_json(to_json(p))


# def test_dodged():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/bars/dodged.html"""
#     from_json(to_json(p))


# def test_stacked_area():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/areas/stacked_area.html"""
#     from_json(to_json(p))


# def test_anscombe():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/layouts/anscombe.html"""
#     from_json(to_json(p))


# def test_legend():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/legend.html"""
#     from_json(to_json(p))


# def test_arrow():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/arrow.html"""
#     from_json(to_json(p))


# def test_band():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/band.html"""
#     from_json(to_json(p))


# def test_slope():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/slope.html"""
#     from_json(to_json(p))


# def test_span():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/span.html"""
#     from_json(to_json(p))


# def test_whisker():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/whisker.html"""
#     from_json(to_json(p))


# def test_colorbar_log():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/colorbar_log.html"""
#     from_json(to_json(p))


# def test_box_annotation():
#     """https://docs.bokeh.org/en/latest/docs/examples/basic/annotations/box_annotation.html"""
#     from_json(to_json(p))


# def test_grid_bounds():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/grid_bounds.html"""
#     from_json(to_json(p))


# def test_minor_grid_lines():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/minor_grid_lines.html"""
#     from_json(to_json(p))


# def test_grid_band_fill():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/grid_band_fill.html"""
#     from_json(to_json(p))


# def test_hatch_grid_band():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/hatch_grid_band.html"""
#     from_json(to_json(p))


# def test_glyph_selection():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/glyph_selection.html"""
#     from_json(to_json(p))


# def test_glyph_hover():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/glyph_hover.html"""
#     from_json(to_json(p))


# def test_legend_location_outside():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/legend_location_outside.html"""
#     from_json(to_json(p))


# def test_legend_title():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/plots/legend_title.html"""
#     from_json(to_json(p))


# def test_latex_blackbody_radiation():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/mathtext/latex_blackbody_radiation.html"""
#     from_json(to_json(p))


# def test_latex_normal_distribution():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/mathtext/latex_normal_distribution.html"""
#     from_json(to_json(p))


# def test_latex_schrodinger():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/mathtext/latex_schrodinger.html"""
#     from_json(to_json(p))


def test_latex_bessel():
    """https://docs.bokeh.org/en/latest/docs/examples/styling/mathtext/latex_bessel.html"""
    p = figure(
        width=700,
        height=500,
        title=r"$$\color{white} \text{Bessel functions of the first kind: } J_\alpha(x) = \sum_{m=0}^{\infty}"
        r"\frac{(-1)^m}{m!\:\Gamma(m+\alpha+1)} \left(\frac{x}{2}\right)^{2m+\alpha}$$",
    )
    p.x_range.range_padding = 0
    p.xaxis.axis_label = r"$$\color{white} x$$"
    p.yaxis.axis_label = r"$$\color{white} J_\alpha(x)$$"
    p.title.text_font_size = "14px"
    x = np.linspace(0.0, 14.0, 100)
    for i, (xlabel, ylabel) in enumerate(
        zip([0.5, 1.6, 2.8, 4.2], [0.95, 0.6, 0.5, 0.45])
    ):
        p.line(x, jv(i, x), line_width=3, color=YlOrRd4[i])
        p.add_layout(
            Label(
                text=r"$$\color{white} J_" + str(i) + "(x)$$",
                x=xlabel,
                y=ylabel,
            )
        )
    from_json(to_json(p))


# def test_caliber():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/themes/caliber.html"""
#     from_json(to_json(p))


# def test_dark_minimal():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/themes/dark_minimal.html"""
#     from_json(to_json(p))


# def test_light_minimal():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/themes/light_minimal.html"""
#     from_json(to_json(p))


# def test_night_sky():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/themes/night_sky.html"""
#     from_json(to_json(p))


# def test_contrast():
#     """https://docs.bokeh.org/en/latest/docs/examples/styling/themes/contrast.html"""
#     from_json(to_json(p))


# def test_image_rgba():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/images/image_rgba.html"""
#     from_json(to_json(p))


# def test_image():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/images/image.html"""
#     from_json(to_json(p))


# def test_image_origin_anchor():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/images/image_origin_anchor.html"""
#     from_json(to_json(p))


def test_contour_simple():
    """https://docs.bokeh.org/en/latest/docs/examples/topics/contour/contour_simple.html"""
    x, y = np.meshgrid(np.linspace(0, 3, 40), np.linspace(0, 2, 30))
    z = 1.3 * np.exp(-2.5 * ((x - 1.3) ** 2 + (y - 0.8) ** 2)) - 1.2 * np.exp(
        -2 * ((x - 1.8) ** 2 + (y - 1.3) ** 2)
    )
    p = figure(width=550, height=300, x_range=(0, 3), y_range=(0, 2))
    levels = np.linspace(-1, 1, 9)
    contour_renderer = p.contour(
        x, y, z, levels, fill_color=Sunset8, line_color="black"
    )
    colorbar = contour_renderer.construct_color_bar()
    p.add_layout(colorbar, "right")
    from_json(to_json(p))


# def test_contour():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/contour/contour.html"""
#     from_json(to_json(p))


# def test_contour_polar():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/contour/contour_polar.html"""
#     from_json(to_json(p))


# def test_hex_tile():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/hex/hex_tile.html"""
#     from_json(to_json(p))


# def test_hexbin():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/hex/hexbin.html"""
#     from_json(to_json(p))


# def test_ridgeplot():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/categorical/ridgeplot.html"""
#     from_json(to_json(p))


# def test_scatter_jitter():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/categorical/scatter_jitter.html"""
#     from_json(to_json(p))


# def test_les_mis():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/categorical/les_mis.html"""
#     from_json(to_json(p))


# def test_heatmap_unemployment():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/categorical/heatmap_unemployment.html"""
#     from_json(to_json(p))


# def test_periodic():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/categorical/periodic.html"""
#     from_json(to_json(p))


# def test_treemap():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/hierarchical/treemap.html"""
#     from_json(to_json(p))


# def test_crosstab():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/hierarchical/crosstab.html"""
#     from_json(to_json(p))


def test_texas_hover_map():
    """https://docs.bokeh.org/en/latest/docs/examples/topics/geo/texas_hover_map.html"""
    from bokeh.sampledata.unemployment import data as unemployment
    from bokeh.sampledata.us_counties import data as counties

    palette = tuple(reversed(Viridis6))
    counties = {
        code: county
        for code, county in counties.items()
        if county["state"] == "tx"
    }
    county_xs = [county["lons"] for county in counties.values()]
    county_ys = [county["lats"] for county in counties.values()]
    county_names = [county["name"] for county in counties.values()]
    county_rates = [unemployment[county_id] for county_id in counties]
    color_mapper = LogColorMapper(palette=palette)
    data = dict(
        x=county_xs,
        y=county_ys,
        name=county_names,
        rate=county_rates,
    )
    TOOLS = "pan,wheel_zoom,reset,hover,save"
    p = figure(
        title="Texas Unemployment, 2009",
        tools=TOOLS,
        x_axis_location=None,
        y_axis_location=None,
        tooltips=[
            ("Name", "@name"),
            ("Unemployment rate", "@rate%"),
            ("(Long, Lat)", "($x, $y)"),
        ],
    )
    p.grid.grid_line_color = None
    p.hover.point_policy = "follow_mouse"
    p.patches(
        "x",
        "y",
        source=data,
        fill_color={"field": "rate", "transform": color_mapper},
        fill_alpha=0.7,
        line_color="white",
        line_width=0.5,
    )
    from_json(to_json(p))


# def test_eclipse():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/geo/eclipse.html"""
#     from_json(to_json(p))


# def test_tile_source():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/geo/tile_source.html"""
#     from_json(to_json(p))


# def test_tile_xyzservices():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/geo/tile_xyzservices.html"""
#     from_json(to_json(p))


# def test_tile_demo():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/geo/tile_demo.html"""
#     from_json(to_json(p))


# def test_gmap():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/geo/gmap.html"""
#     from_json(to_json(p))


# def test_from_networkx():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/graph/from_networkx.html"""
#     from_json(to_json(p))


# def test_node_and_edge_attributes():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/graph/node_and_edge_attributes.html"""
#     from_json(to_json(p))


# def test_candlestick():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/timeseries/candlestick.html"""
#     from_json(to_json(p))


# def test_missing_dates():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/timeseries/missing_dates.html"""
#     from_json(to_json(p))


# def test_pie():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/pie/pie.html"""
#     from_json(to_json(p))


# def test_donut():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/pie/donut.html"""
#     from_json(to_json(p))


# def test_burtin():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/pie/burtin.html"""
#     from_json(to_json(p))


# def test_histogram():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/stats/histogram.html"""
#     from_json(to_json(p))


def test_kde2d():
    """https://docs.bokeh.org/en/latest/docs/examples/topics/stats/kde2d.html"""
    from bokeh.sampledata.autompg import autompg as df

    def kde(x, y, N):
        xmin, xmax = x.min(), x.max()
        ymin, ymax = y.min(), y.max()
        X, Y = np.mgrid[xmin : xmax : N * 1j, ymin : ymax : N * 1j]
        positions = np.vstack([X.ravel(), Y.ravel()])
        values = np.vstack([x, y])
        kernel = gaussian_kde(values)
        Z = np.reshape(kernel(positions).T, X.shape)
        return X, Y, Z

    x, y, z = kde(df.hp, df.mpg, 300)
    p = figure(
        height=400,
        x_axis_label="hp",
        y_axis_label="mpg",
        background_fill_color="#fafafa",
        tools="",
        toolbar_location=None,
        title="Kernel density estimation plot of HP vs MPG",
    )
    p.grid.level = "overlay"
    p.grid.grid_line_color = "black"
    p.grid.grid_line_alpha = 0.05
    palette = Blues9[::-1]
    levels = np.linspace(np.min(z), np.max(z), 10)
    p.contour(x, y, z, levels[1:], fill_color=palette, line_color=palette)
    from_json(to_json(p))


# def test_splom():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/stats/splom.html"""
#     from_json(to_json(p))


# def test_boxplot():
#     """https://docs.bokeh.org/en/latest/docs/examples/topics/stats/boxplot.html"""
#     from_json(to_json(p))


def test_range_tool():
    """https://docs.bokeh.org/en/latest/docs/examples/interaction/tools/range_tool.html"""
    from bokeh.sampledata.stocks import AAPL

    dates = np.array(AAPL["date"], dtype=np.datetime64)
    source = ColumnDataSource(data=dict(date=dates, close=AAPL["adj_close"]))
    p = figure(
        height=300,
        width=800,
        tools="xpan",
        toolbar_location=None,
        x_axis_type="datetime",
        x_axis_location="above",
        background_fill_color="#efefef",
        x_range=(dates[1500], dates[2500]),
    )
    p.line("date", "close", source=source)
    p.yaxis.axis_label = "Price"
    select = figure(
        title="Drag the middle and edges of the selection box to change the range above",
        height=130,
        width=800,
        y_range=p.y_range,
        x_axis_type="datetime",
        y_axis_type=None,
        tools="",
        toolbar_location=None,
        background_fill_color="#efefef",
    )
    range_tool = RangeTool(x_range=p.x_range)
    range_tool.overlay.fill_color = "navy"
    range_tool.overlay.fill_alpha = 0.2
    select.line("date", "close", source=source)
    select.ygrid.grid_line_color = None
    select.add_tools(range_tool)
    p = column(p, select)
    from_json(to_json(p))


# def test_linked_brushing():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/linking/linked_brushing.html"""
#     from_json(to_json(p))


# def test_linked_crosshair():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/linking/linked_crosshair.html"""
#     from_json(to_json(p))


# def test_data_table_plot():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/linking/data_table_plot.html"""
#     from_json(to_json(p))


# def test_legend_hide():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/legends/legend_hide.html"""
#     from_json(to_json(p))


# def test_legend_mute():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/legends/legend_mute.html"""
#     from_json(to_json(p))


def test_slider():
    """https://docs.bokeh.org/en/latest/docs/examples/interaction/js_callbacks/slider.html"""
    x = np.linspace(0, 10, 500)
    y = np.sin(x)
    source = ColumnDataSource(data=dict(x=x, y=y))
    plot = figure(y_range=(-10, 10), width=400, height=400)
    plot.line("x", "y", source=source, line_width=3, line_alpha=0.6)
    amp = Slider(start=0.1, end=10, value=1, step=0.1, title="Amplitude")
    freq = Slider(start=0.1, end=10, value=1, step=0.1, title="Frequency")
    phase = Slider(start=-6.4, end=6.4, value=0, step=0.1, title="Phase")
    offset = Slider(start=-9, end=9, value=0, step=0.1, title="Offset")
    callback = CustomJS(
        args=dict(
            source=source, amp=amp, freq=freq, phase=phase, offset=offset
        ),
        code="""
        const A = amp.value
        const k = freq.value
        const phi = phase.value
        const B = offset.value

        const x = source.data.x
        const y = Array.from(x, (x) => B + A*Math.sin(k*x+phi))
        source.data = { x, y }
    """,
    )
    amp.js_on_change("value", callback)
    freq.js_on_change("value", callback)
    phase.js_on_change("value", callback)
    offset.js_on_change("value", callback)
    p = row(plot, column(amp, freq, phase, offset))
    from_json(to_json(p))


# def test_color_sliders():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/js_callbacks/color_sliders.html"""
#     from_json(to_json(p))


def test_customjs_lasso_mean():
    """https://docs.bokeh.org/en/latest/docs/examples/interaction/js_callbacks/customjs_lasso_mean.html"""
    x = [random() for x in range(500)]
    y = [random() for y in range(500)]
    color = ["navy"] * len(x)
    s = ColumnDataSource(data=dict(x=x, y=y, color=color))
    p = figure(
        width=400, height=400, tools="lasso_select", title="Select Here"
    )
    p.circle(
        "x",
        "y",
        color="color",
        size=8,
        source=s,
        alpha=0.4,
        selection_color="firebrick",
    )
    s2 = ColumnDataSource(data=dict(x=[0, 1], ym=[0.5, 0.5]))
    p.line(x="x", y="ym", color="orange", line_width=5, alpha=0.6, source=s2)
    s.selected.js_on_change(
        "indices",
        CustomJS(
            args=dict(s=s, s2=s2),
            code="""
        const inds = s.selected.indices
        if (inds.length > 0) {
            const ym = inds.reduce((a, b) => a + s.data.y[b], 0) / inds.length
            s2.data = { x: s2.data.x, ym: [ym, ym] }
        }
    """,
        ),
    )
    from_json(to_json(p))


# def test_js_on_event():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/js_callbacks/js_on_event.html"""
#     from_json(to_json(p))


# def test_multiselect():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/widgets/multiselect.html"""
#     from_json(to_json(p))


# def test_multichoice():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/widgets/multichoice.html"""
#     from_json(to_json(p))


# def test_date_picker():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/widgets/date_picker.html"""
#     from_json(to_json(p))


# def test_dropdown():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/widgets/dropdown.html"""
#     from_json(to_json(p))


# def test_data_table():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/widgets/data_table.html"""
#     from_json(to_json(p))


# def test_data_cube():
#     """https://docs.bokeh.org/en/latest/docs/examples/interaction/widgets/data_cube.html"""
#     from_json(to_json(p))



"""Scikit-learn estimators"""
__docformat__ = "google"

import re
from typing import Any, Callable, Dict, List, Tuple

from sklearn import (
    calibration,
    cluster,
    compose,
    covariance,
    cross_decomposition,
    datasets,
    decomposition,
    discriminant_analysis,
    dummy,
    ensemble,
    exceptions,
    feature_extraction,
    feature_selection,
    gaussian_process,
    impute,
    inspection,
    isotonic,
    kernel_approximation,
    kernel_ridge,
    linear_model,
    manifold,
    metrics,
    mixture,
    model_selection,
    multiclass,
    multioutput,
    naive_bayes,
    neighbors,
    neural_network,
    pipeline,
    preprocessing,
    random_projection,
    semi_supervised,
    svm,
    tree,
)
from sklearn.base import BaseEstimator
from sklearn.tree._tree import Tree

_SKLEARN_SUBMODULES = [
    # calibration,
    cluster,
    covariance,
    cross_decomposition,
    datasets,
    decomposition,
    # dummy,
    ensemble,
    exceptions,
    # experimental,
    # externals,
    feature_extraction,
    feature_selection,
    gaussian_process,
    inspection,
    isotonic,
    # kernel_approximation,
    # kernel_ridge,
    linear_model,
    manifold,
    metrics,
    mixture,
    model_selection,
    multiclass,
    multioutput,
    naive_bayes,
    neighbors,
    neural_network,
    pipeline,
    preprocessing,
    random_projection,
    semi_supervised,
    svm,
    tree,
    discriminant_analysis,
    impute,
    compose,
]

_SKLEARN_TREE_ATTRIBUTES = [
    "capacity",
    "children_left",
    "children_right",
    "feature",
    "impurity",
    "max_depth",
    "max_n_classes",
    "n_classes",
    "n_features",
    "n_leaves",
    "n_node_samples",
    "n_outputs",
    "node_count",
    "threshold",
    "value",
    "weighted_n_node_samples",
]


def _all_base_estimators() -> Dict[str, type]:
    """
    Returns (hopefully) all classes of sklearn that inherit from
    `BaseEstimator`
    """
    result = []
    for s in _SKLEARN_SUBMODULES:
        if not hasattr(s, "__all__"):
            continue
        s_all = getattr(s, "__all__")
        if not isinstance(s_all, list):
            continue
        for k in s_all:
            cls = getattr(s, k)
            if isinstance(cls, type) and issubclass(cls, BaseEstimator):
                result.append(cls)
    # Some sklearn submodules don't have __all__
    result += [
        calibration.CalibratedClassifierCV,
        dummy.DummyClassifier,
        dummy.DummyRegressor,
        kernel_approximation.PolynomialCountSketch,
        kernel_approximation.RBFSampler,
        kernel_approximation.SkewedChi2Sampler,
        kernel_approximation.AdditiveChi2Sampler,
        kernel_approximation.Nystroem,
        kernel_ridge.KernelRidge,
    ]
    return {cls.__name__: cls for cls in result}


def _sklearn_estimator_to_json(obj: BaseEstimator) -> dict:
    """Converts a sklearn estimator into a JSON document."""
    r = re.compile(r"\w[\w_]*[^_]_")
    return {
        "__type__": "estimator",
        "cls": obj.__class__.__name__,
        "__version__": 1,
        "params": obj.get_params(deep=False),
        "attrs": {k: v for k, v in obj.__dict__.items() if r.match(k)},
    }


def _sklearn_tree_to_json(obj: Tree) -> dict:
    """Converts a sklearn Tree object into a JSON document."""
    return {
        "__type__": "tree",
        "__version__": 1,
        **{a: getattr(obj, a) for a in _SKLEARN_TREE_ATTRIBUTES},
    }


def _json_to_sklearn_estimator(dct: dict) -> BaseEstimator:
    """
    Converts a JSON document to a sklearn estimator. See `to_json` for the
    specification `dct` is expected to follow. Note that the key
    `__sklearn__` should not be present.
    """
    DECODERS = {
        1: _json_to_sklearn_estimator_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_sklearn_estimator_v1(dct: dict) -> BaseEstimator:
    """
    Converts a JSON document to a sklearn estimator following the v1
    specification.
    """
    bes = _all_base_estimators()
    cls = bes[dct["cls"]]
    obj = cls(**dct["params"])
    for k, v in dct["attrs"].items():
        setattr(obj, k, v)
    return obj


def _json_to_sklearn_tree(dct: dict) -> BaseEstimator:
    """
    Converts a JSON document to a sklearn tree object. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__sklearn__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_sklearn_tree_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_sklearn_tree_v1(dct: dict) -> BaseEstimator:
    """
    Converts a JSON document to a sklearn tree following the v1 specification.

    See also:
        * https://github.com/scikit-learn/scikit-learn/blob/9aaed498795f68e5956ea762fef9c440ca9eb239/sklearn/tree/_tree.pxd
        * https://github.com/scikit-learn/scikit-learn/blob/9aaed498795f68e5956ea762fef9c440ca9eb239/sklearn/tree/_classes.py#L349
    """
    obj = Tree(dct["n_features"], dct["n_classes"], dct["n_outputs"])
    for k in _SKLEARN_TREE_ATTRIBUTES:
        setattr(obj, k, dct[k])
    return obj


def from_json(dct: dict) -> BaseEstimator:
    """
    Deserializes a dict into a sklearn estimator. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__sklearn__`.
    """
    DECODERS = {
        "estimator": _json_to_sklearn_estimator,
        "tree": _json_to_sklearn_tree,
    }
    try:
        return DECODERS[dct["__sklearn__"]["__type__"]](dct["__sklearn__"])
    except KeyError as exc:
        raise TypeError("Not a valid sklearn document") from exc


def to_json(obj: BaseEstimator) -> dict:
    """
    Serializes a sklearn estimator into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__sklearn__": {
                "__type__": "estimator",
                "__version__": 1,
                "cls": <class name>,
                "params": <dict returned by get_params(deep=False)>,
                "attrs": {...}
            },
        }

    where the `attrs` dict contains all the attributes of the estimator as
    specified in the sklearn API documentation.
    """
    ENCODERS: List[Tuple[type, Callable[[Any], dict]]] = [
        (BaseEstimator, _sklearn_estimator_to_json),
        (Tree, _sklearn_tree_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__sklearn__": f(obj)}
    raise TypeError("Not a supported sklearn type")



"""scipy objects"""
__docformat__ = "google"

from typing import Any, Callable, List, Tuple

from scipy.sparse import csr_matrix


def _csr_matrix_to_json(m: csr_matrix) -> dict:
    """Converts a csr_matrix into a JSON document."""
    return {
        "__type__": "csr_matrix",
        "__version__": 1,
        "data": m.data,
        "dtype": m.dtype,
        "indices": m.indices,
        "indptr": m.indptr,
        "shape": m.shape,
    }


def _json_to_csr_matrix(dct: dict) -> csr_matrix:
    """
    Converts a JSON document to a csr_matrix. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__scipy__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_csr_matrix_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_csr_matrix_v1(dct: dict) -> csr_matrix:
    """
    Converts a JSON document to a csr_matrix following the v1 specification.
    """
    return csr_matrix(
        (dct["data"], dct["indices"], dct["indptr"]),
        shape=dct["shape"],
        dtype=dct["dtype"],
    )


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a csr_matrix. See `to_json` for the specification
    `dct` is expected to follow. In particular, note that `dct` must contain
    the key `__csr_matrix__`.
    """
    DECODERS = {
        "csr_matrix": _json_to_csr_matrix,
    }
    try:
        return DECODERS[dct["__scipy__"]["__type__"]](dct["__scipy__"])
    except KeyError as exc:
        raise TypeError("Not a valid scipy document") from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a Scipy object into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__scipy__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.
    """
    ENCODERS: List[Tuple[type, Callable[[Any], dict]]] = [
        (csr_matrix, _csr_matrix_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__scipy__": f(obj)}
    raise TypeError("Not a supported scipy type")



# pylint: disable=missing-function-docstring
"""Numpy (de)serialization test suite"""

from datetime import datetime

import pandas as pd

from common import from_json, to_json


def _assert_equal(a, b):
    assert type(a) == type(b)  # pylint: disable=unidiomatic-typecheck
    if isinstance(a, pd.DataFrame):
        assert ((a == b).all()).all()
    elif isinstance(a, pd.Series):
        assert (a == b).all()


def test_series_int():
    x = pd.Series(range(10))
    _assert_equal(x, from_json(to_json(x)))
    x = pd.Series(range(10), name="a_series")
    _assert_equal(x, from_json(to_json(x)))


def test_series_float():
    x = pd.Series([1.0, 0.1, 1e3], name="a_series")
    _assert_equal(x, from_json(to_json(x)))


# def test_series_complex():
#     x = pd.Series([1.0, 0.1j, 1e3 + 1e4j], name="a_series")
#     _assert_equal(x, from_json(to_json(x)))


def test_series_string():
    x = pd.Series(["a", "b", "c", "🧇"], name="a_series")
    _assert_equal(x, from_json(to_json(x)))


def test_series_datetime():
    x = pd.Series(
        ["2013-01-01", "2013-01-02", "2013-01-03", datetime.now()],
        dtype="datetime64[ns]",
        name="a_series",
    )
    _assert_equal(x, from_json(to_json(x)))


def test_series_categorical():
    x = pd.Series(
        ["A", "B", "A", "C", "B", "A"],
        dtype="category",
        name="a_series",
    )
    _assert_equal(x, from_json(to_json(x)))


def test_dataframe():
    x = pd.DataFrame(
        {
            "float": pd.Series(["1", ".1", "1e-10", "1e11"], dtype="float64"),
            "date": datetime.now(),
            # "td": pd.Timedelta("1 day"),
            "string": "Hello world 👋",
            # "cplx": [1, 2 + 3j, 3.14, .0000001],
            "cat": pd.Categorical(["X", "Y", "X", "Z"]),
        },
    )
    _assert_equal(x, from_json(to_json(x)))



"""pandas (de)serialization utilities."""
__docformat__ = "google"

import json
from typing import Any, Callable, List, Tuple
from uuid import uuid4

import pandas as pd
from turbo_broccoli.environment import (
    get_artifact_path,
    get_max_nbytes,
    get_pandas_format,
    is_nodecode,
)


def _dataframe_to_json(df: pd.DataFrame) -> dict:
    """Converts a pandas dataframe into a JSON document."""
    # Sometimes column names are int, so cannot be used as keys in a JSON
    # document. Eventhough int column names are not supported, this is
    # future-proofing.
    dtypes = [(k, v.name) for k, v in df.dtypes.items()]
    if df.memory_usage(deep=True).sum() <= get_max_nbytes():
        return {
            "__type__": "dataframe",
            "__version__": 1,
            "data": json.loads(df.to_json(date_format="iso", date_unit="ns")),
            "dtypes": dtypes,
        }
    fmt = get_pandas_format()
    name = str(uuid4())
    path = get_artifact_path() / name
    if fmt in ["h5", "hdf"]:
        df.to_hdf(path, "main")
    else:
        getattr(df, f"to_{fmt}")(path)
    return {
        "__type__": "dataframe",
        "__version__": 1,
        "dtypes": dtypes,
        "id": name,
        "format": fmt,
    }


def _json_to_dataframe(dct: dict) -> pd.DataFrame:
    """
    Converts a JSON document to a pandas dataframe. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__pandas__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_dataframe_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_dataframe_v1(dct: dict) -> pd.DataFrame:
    """
    Converts a JSON document to a pandas dataframe following the v1
    specification.
    """
    if "data" in dct:
        df = pd.read_json(json.dumps(dct["data"]))
    else:
        fmt = dct["format"]
        path = get_artifact_path() / dct["id"]
        if fmt in ["h5", "hdf"]:
            df = pd.read_hdf(path, "main")
        else:
            df = getattr(pd, f"read_{fmt}")(path)
    # Rename columns with non-string names
    # df.rename({str(d[0]): d[0] for d in dct["dtypes"]}, inplace=True)
    df = df.astype(
        {
            str(a): b
            for a, b in dct["dtypes"]
            if not str(b).startswith("datetime")
        }
    )
    for a, _ in filter(lambda x: x[1].startswith("datetime"), dct["dtypes"]):
        df[a] = pd.to_datetime(df[a]).dt.tz_localize(None)
    return df


def _json_to_series(dct: dict) -> pd.Series:
    """
    Converts a JSON document to a pandas series. See `to_json` for the
    specification `dct` is expected to follow. Note that the key `__pandas__`
    should not be present.
    """
    DECODERS = {
        1: _json_to_series_v1,
    }
    return DECODERS[dct["__version__"]](dct)


def _json_to_series_v1(dct: dict) -> pd.Series:
    """
    Converts a JSON document to a pandas series following the v1 specification.
    """
    return dct["data"][dct["name"]]


def _series_to_json(ser: pd.Series) -> dict:
    """Converts a pandas series into a JSON document."""
    name = ser.name if ser.name is not None else "main"
    return {
        "__type__": "series",
        "__version__": 1,
        "data": ser.to_frame(name=name),
        "name": name,
    }


def from_json(dct: dict) -> Any:
    """
    Deserializes a dict into a pandas object. See `to_json` for the
    specification `dct` is expected to follow. In particular, note that `dct`
    must contain the key `__pandas__`.
    """
    DECODERS = {
        "dataframe": _json_to_dataframe,
        "series": _json_to_series,
    }
    try:
        type_name = dct["__pandas__"]["__type__"]
        if is_nodecode("pandas." + type_name):
            return None
        return DECODERS[type_name](dct["__pandas__"])
    except KeyError as exc:
        raise TypeError("Not a valid pandas document") from exc


def to_json(obj: Any) -> dict:
    """
    Serializes a pandas object into JSON by cases. See the README for the
    precise list of supported types.

    The return dict has the following structure

        {
            "__pandas__": {...},
        }

    where the `{...}` dict contains the actual data, and whose structure
    depends on the precise type of `obj`.

    * `pandas.DataFrame`: A dataframe is processed differently depending on its
      size and on the `TB_MAX_NBYTES` environment variable. If the dataframe is
      small, i.e. at most `TB_MAX_NBYTES` bytes, then it is directly stored in
      the resulting JSON document as

            {
                "__pandas__": {
                    "__type__": "dataframe",
                    "__version__": 1,
                    "data": {...},
                    "dtypes": [
                        [col1, dtype1],
                        [col2, dtype2],
                        ...
                    ],
                }
            }

      where `{...}` is the result of `pandas.DataFrame.to_json` (in `dict`
      form). On the other hand, the dataframe is too but, then its content is
      stored in an artifact, whose format follows the `TB_PANDAS_FORMAT`
      environment (HDF5 by default). Said file is saved to the
      path specified by the `TB_ARTIFACT_PATH` environment variable with a
      random UUID4 as filename. The resulting JSON document looks like

            {
                "__pandas__": {
                    "__type__": "dataframe",
                    "__version__": 1,
                    "dtypes": [
                        [col1, dtype1],
                        [col2, dtype2],
                        ...
                    ],
                    "id": <UUID4 str>,
                    "format": <str>
                }
            }

    * `pandas.Series`: A series will be converted to a dataframe before being
      serialized. The final document will look like this

            {
                "__pandas__": {
                    "__type__": "series",
                    "__version__": 1,
                    "data": {...},
                    "name": <str>,
                }
            }

      where `{...}` is the document of the dataframe'd series. So for example

            {
                "__pandas__": {
                    "__type__": "series",
                    "__version__": 1,
                    "data": {
                        "__type__": "dataframe",
                        "__version__": 1,
                        "id": <UUID4 str>,
                        "format": <str>,
                    },
                    "name": <str>,
                }
            }

      if the series is large.

    Warning:
        Series and column names must be strings!

    """
    ENCODERS: List[Tuple[type, Callable[[Any], dict]]] = [
        (pd.DataFrame, _dataframe_to_json),
        (pd.Series, _series_to_json),
    ]
    for t, f in ENCODERS:
        if isinstance(obj, t):
            return {"__pandas__": f(obj)}
    raise TypeError("Not a supported pandas type")



"""Main module containing the JSON encoder and decoder classes."""
__docformat__ = "google"

import json
from typing import Any, Callable, Dict, List


import turbo_broccoli.bytes
import turbo_broccoli.collections
import turbo_broccoli.dataclass
import turbo_broccoli.generic

try:
    import turbo_broccoli.keras

    HAS_KERAS = True
except ModuleNotFoundError:
    HAS_KERAS = False

try:
    import turbo_broccoli.numpy

    HAS_NUMPY = True
except ModuleNotFoundError:
    HAS_NUMPY = False

try:
    import turbo_broccoli.pandas

    HAS_PANDAS = True
except ModuleNotFoundError:
    HAS_PANDAS = False


try:
    import turbo_broccoli.secret

    HAS_SECRET = True
except ModuleNotFoundError:
    HAS_SECRET = False

try:
    import turbo_broccoli.tensorflow

    HAS_TENSORFLOW = True
except ModuleNotFoundError:
    HAS_TENSORFLOW = False

try:
    import turbo_broccoli.pytorch

    HAS_PYTORCH = True
except ModuleNotFoundError:
    HAS_PYTORCH = False


try:
    import turbo_broccoli.scipy

    HAS_SCIPY = True
except ModuleNotFoundError:
    HAS_SCIPY = False

try:
    import turbo_broccoli.sklearn

    HAS_SKLEARN = True
except ModuleNotFoundError:
    HAS_SKLEARN = False


try:
    import turbo_broccoli.bokeh

    HAS_BOKEH = True
except ModuleNotFoundError:
    HAS_BOKEH = False


class TurboBroccoliDecoder(json.JSONDecoder):
    """
    TurboBroccoli's custom JSON decoder class. See the README for the list of
    supported types.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(object_hook=self._hook, *args, **kwargs)

    def _hook(self, dct):
        """Deserialization hook"""
        DECODERS: Dict[str, Callable[[dict], Any]] = {
            "__bytes__": turbo_broccoli.bytes.from_json,
        }
        if HAS_KERAS:
            DECODERS["__keras__"] = turbo_broccoli.keras.from_json
        if HAS_NUMPY:
            DECODERS["__numpy__"] = turbo_broccoli.numpy.from_json
        if HAS_PANDAS:
            DECODERS["__pandas__"] = turbo_broccoli.pandas.from_json
        if HAS_PYTORCH:
            DECODERS["__pytorch__"] = turbo_broccoli.pytorch.from_json
        if HAS_SECRET:
            DECODERS["__secret__"] = turbo_broccoli.secret.from_json
        if HAS_TENSORFLOW:
            DECODERS["__tensorflow__"] = turbo_broccoli.tensorflow.from_json
        if HAS_SCIPY:
            DECODERS["__scipy__"] = turbo_broccoli.scipy.from_json
        if HAS_SKLEARN:
            DECODERS["__sklearn__"] = turbo_broccoli.sklearn.from_json
        if HAS_BOKEH:
            DECODERS["__bokeh__"] = turbo_broccoli.bokeh.from_json
        # Intentionally put last
        DECODERS["__collections__"] = turbo_broccoli.collections.from_json
        DECODERS["__dataclass__"] = turbo_broccoli.dataclass.from_json
        for t, f in DECODERS.items():
            if t in dct:
                return f(dct)
        return dct


class TurboBroccoliEncoder(json.JSONEncoder):
    """
    TurboBroccoli's custom JSON decoder class. See the README for the list of
    supported types.
    """

    def default(self, o: Any) -> Any:

        ENCODERS: List[Callable[[Any], dict]] = [
            turbo_broccoli.bytes.to_json,
        ]
        if HAS_KERAS:
            ENCODERS.append(turbo_broccoli.keras.to_json)
        if HAS_NUMPY:
            ENCODERS.append(turbo_broccoli.numpy.to_json)
        if HAS_PANDAS:
            ENCODERS.append(turbo_broccoli.pandas.to_json)
        if HAS_PYTORCH:
            ENCODERS.append(turbo_broccoli.pytorch.to_json)
        if HAS_SECRET:
            ENCODERS.append(turbo_broccoli.secret.to_json)
        if HAS_TENSORFLOW:
            ENCODERS.append(turbo_broccoli.tensorflow.to_json)
        if HAS_SCIPY:
            ENCODERS.append(turbo_broccoli.scipy.to_json)
        if HAS_SKLEARN:
            ENCODERS.append(turbo_broccoli.sklearn.to_json)
        if HAS_BOKEH:
            ENCODERS.append(turbo_broccoli.bokeh.to_json)
        # Intentionally put last
        ENCODERS += [
            turbo_broccoli.collections.to_json,
            turbo_broccoli.dataclass.to_json,
            turbo_broccoli.generic.to_json,
        ]
        for f in ENCODERS:
            try:
                return f(o)
            except TypeError:
                pass
        return super().default(o)

    def encode(self, o: Any) -> str:
        """
        Reimplementation of encode just to treat the `namedtuple` case. An
        object is considered a namedtuple if it is an instance of `tuple` and
        has the following attributes: `_asdict`, `_field_defaults`, `_fields`,
        `_make`, `_replace`. In this case,
        `turbo_broccoli.collections._namedtuple_to_json` is called directly.
        """
        attrs = ["_asdict", "_field_defaults", "_fields", "_make", "_replace"]
        if isinstance(o, tuple) and all(map(lambda a: hasattr(o, a), attrs)):
            d = turbo_broccoli.collections._namedtuple_to_json(o)
            return super().encode({"__collections__": d})
        return super().encode(o)


def from_json(doc: str) -> Any:
    """Converts a JSON document back to a Python object."""
    return json.loads(doc, cls=TurboBroccoliDecoder)


def to_json(obj: Any) -> str:
    """Converts an object to JSON."""
    return json.dumps(obj, cls=TurboBroccoliEncoder)
